import os

import pytest
from itsdangerous import BadSignature

from blacksheep.server.dataprotection import generate_secret, get_keys, get_serializer


def test_get_keys_creates_default_keys():
    default_keys = get_keys()

    assert default_keys is not None
    assert len(default_keys) == 3

    assert default_keys != get_keys()


def test_get_keys_returns_keys_configured_as_env_variables():
    env_variables = []
    for i in range(4):
        key = generate_secret()
        os.environ[f"APP_SECRET_{i}"] = key
        env_variables.append(key)

    assert get_keys() == env_variables
    assert get_keys() == env_variables


def test_get_serializer():
    serializer = get_serializer(get_keys())

    data = {"id": "0000"}
    secret = serializer.dumps(data)
    assert isinstance(secret, str)

    parsed = serializer.loads(secret)
    assert data == parsed


def test_get_serializer_with_different_purpose():
    keys = get_keys()
    serializer = get_serializer(keys)
    other_serializer = get_serializer(keys, purpose="test")

    data = {"id": "0000"}
    secret = serializer.dumps(data)

    with pytest.raises(BadSignature):
        other_serializer.loads(secret)


def test_get_serializer_with_default_keys():
    serializer = get_serializer()

    data = {"id": "0000"}
    secret = serializer.dumps(data)
    assert isinstance(secret, str)

    parsed = serializer.loads(secret)
    assert data == parsed



import abc
from typing import Dict, Optional

from blacksheep.contents import Content
from blacksheep.messages import Request
from blacksheep.server.application import Application
from blacksheep.server.responses import Response
from blacksheep.testing.helpers import get_example_scope

from .helpers import CookiesType, HeadersType, QueryType


def _create_scope(
    method: str,
    path: str,
    headers: HeadersType = None,
    query: QueryType = None,
    cookies: CookiesType = None,
) -> Dict:
    """Creates a mocked ASGI scope"""
    return get_example_scope(
        method, path, extra_headers=headers, query=query, cookies=cookies
    )


class AbstractTestSimulator:
    """An abstract class for custom Test simulator clients"""

    @abc.abstractmethod
    async def send_request(
        self,
        method: str,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Optional[Content] = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Entrypoint for all HTTP methods

        The method is main entrypoint for all TestClient methods
            - get
            - post
            - patch
            - put
            - delete
        Then you can define an own TestClient, with the custom logic.
        """


class TestSimulator(AbstractTestSimulator):
    """Base Test simulator class

    The class introduces a fast "tests" for your server-side application,
    it means that all incoming requests are mocked.
    """

    def __init__(self, app: Application):
        self.app = app
        self._is_started_app()

    async def send_request(
        self,
        method: str,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Optional[Content] = None,
        cookies: CookiesType = None,
    ) -> Response:
        scope = _create_scope(method, path, headers, query, cookies=cookies)
        request = Request.incoming(
            scope["method"],
            scope["raw_path"],
            scope["query_string"],
            scope["headers"],
        )

        request.scope = scope  # type: ignore

        if content is not None:
            if not isinstance(content, Content):
                raise ValueError(
                    "The content argument should be an instance of Content class"
                )

            request.content = content

        response = await self.app.handle(request)

        return response

    def _is_started_app(self):
        assert (
            self.app.started
        ), "The BlackSheep application is not started, use Application.start method"



from mimetypes import MimeTypes

mime = MimeTypes()


DEFAULT_MIME = "application/octet-stream"

MIME_BY_EXTENSION = {
    ".ogg": "audio/ogg",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".png": "image/png",
    ".woff2": "font/woff2",
}


def get_file_extension_from_name(name: str) -> str:
    if not name:
        return ""
    return name[name.rfind(".") :].lower()


def get_mime_type_from_name(file_name: str) -> str:
    extension = get_file_extension_from_name(file_name)
    mime_type = mime.guess_type(file_name)[0] or DEFAULT_MIME

    if mime_type == DEFAULT_MIME:
        mime_type = MIME_BY_EXTENSION.get(extension, DEFAULT_MIME)
    return mime_type



import sys
from dataclasses import dataclass
from datetime import datetime
from io import BytesIO
from typing import Any, AsyncIterable, Callable, Dict, Optional
from uuid import UUID, uuid4

import pytest

from blacksheep import Content, Cookie, Response, scribe
from blacksheep.server.controllers import (
    CannotDetermineDefaultViewNameError,
    Controller,
    TemplatingNotConfiguredException,
)
from blacksheep.server.responses import (
    ContentDispositionType,
    accepted,
    bad_request,
    created,
    file,
    forbidden,
    html,
    json,
    moved_permanently,
    no_content,
    not_found,
    not_modified,
    ok,
    permanent_redirect,
    pretty_json,
    redirect,
    see_other,
    status_code,
    temporary_redirect,
    text,
    unauthorized,
)
from blacksheep.server.routing import RoutesRegistry
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.test_files_serving import get_file_path

STATUS_METHODS_OPTIONAL_BODY = [
    (ok, 200),
    (created, 201),
    (accepted, 202),
    (unauthorized, 401),
    (bad_request, 400),
    (forbidden, 403),
    (not_found, 404),
]

REDIRECT_METHODS = [
    (moved_permanently, 301),
    (redirect, 302),
    (see_other, 303),
    (temporary_redirect, 307),
    (permanent_redirect, 308),
]

STATUS_METHODS_NO_BODY = [(no_content, 204), (not_modified, 304)]

EXAMPLE_HTML = """
<!DOCTYPE html>
<html>
  <head>
    <title>Example</title>
    <style>
      h1 {
          color: pink;
      }
    </style>
    <link rel="stylesheet" type="text/css" href="/home.css" />
  </head>
  <body>
    <h1>Lorem ipsum 🍎</h1>
    <p>Dolor sit amet</p>
  </body>
</html>"""


@dataclass
class Demo:
    id: UUID
    name: str
    created_at: datetime


@dataclass
class Foo:
    id: UUID
    name: str
    ufo: bool


JSON_OBJECTS = [
    (
        Demo(
            UUID("00000000-0000-0000-0000-000000000000"),
            "Foo",
            datetime(2018, 8, 17, 20, 55, 4),
        ),
        {
            "id": "00000000-0000-0000-0000-000000000000",
            "name": "Foo",
            "created_at": "2018-08-17T20:55:04",
        },
    ),
    (
        Demo(
            UUID("00000000-0000-0000-0000-000000000001"),
            "Lorem ipsum",
            datetime(2015, 10, 21, 7, 28, 00),
        ),
        {
            "id": "00000000-0000-0000-0000-000000000001",
            "name": "Lorem ipsum",
            "created_at": "2015-10-21T07:28:00",
        },
    ),
]


async def read_from_asynciterable(method: Callable[[], AsyncIterable[bytes]]) -> str:
    parts = []
    async for chunk in method():
        parts.append(chunk.decode("utf8"))
    return "".join(parts)


def test_response_supports_dynamic_attributes():
    response = Response(200)
    foo = object()

    assert (
        hasattr(response, "response") is False
    ), "This test makes sense if such attribute is not defined"
    response.foo = foo  # type: ignore
    assert response.foo is foo  # type: ignore


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "response,cookies,expected_result",
    [
        (
            Response(400, [(b"Server", b"BlackSheep")]).with_content(
                Content(b"text/plain", b"Hello, World")
            ),
            [],
            b"HTTP/1.1 400 Bad Request\r\n"
            b"Server: BlackSheep\r\n"
            b"content-type: text/plain\r\n"
            b"content-length: 12\r\n\r\nHello, World",
        ),
        (
            Response(400, [(b"Server", b"BlackSheep")]).with_content(
                Content(b"text/plain", b"Hello, World")
            ),
            [Cookie("session", "123")],
            b"HTTP/1.1 400 Bad Request\r\n"
            b"Server: BlackSheep\r\n"
            b"set-cookie: session=123\r\n"
            b"content-type: text/plain\r\n"
            b"content-length: 12\r\n\r\nHello, World",
        ),
        (
            Response(400, [(b"Server", b"BlackSheep")]).with_content(
                Content(b"text/plain", b"Hello, World")
            ),
            [Cookie("session", "123"), Cookie("aaa", "bbb", domain="bezkitu.org")],
            b"HTTP/1.1 400 Bad Request\r\n"
            b"Server: BlackSheep\r\n"
            b"set-cookie: session=123\r\n"
            b"set-cookie: aaa=bbb; Domain=bezkitu.org\r\n"
            b"content-type: text/plain\r\n"
            b"content-length: 12\r\n\r\nHello, World",
        ),
    ],
)
async def test_write_http_response(response, cookies, expected_result):
    response.set_cookies(cookies)
    data = b""
    async for chunk in scribe.write_response(response):
        data += chunk
    assert data == expected_result


def test_is_redirect():
    # 301 Moved Permanently
    # 302 Found
    # 303 See Other
    # 307 Temporary Redirect
    # 308 Permanent Redirect
    for status in range(200, 500):
        response = Response(status)
        is_redirect = status in {301, 302, 303, 307, 308}
        assert response.is_redirect() == is_redirect


@pytest.mark.parametrize("method,expected_status", REDIRECT_METHODS)
@pytest.mark.asyncio
async def test_redirect_method(method, expected_status, app):
    @app.router.get("/")
    async def home():
        return method("https://foo.org/somewhere")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == expected_status
    location = response.headers.get_single(b"location")
    assert location == b"https://foo.org/somewhere"


@pytest.mark.parametrize("method", (method for method, _ in REDIRECT_METHODS))
def test_redirect_method_raises_for_invalid_location(method):
    with pytest.raises(ValueError):
        method(location=True)

    with pytest.raises(ValueError):
        method(location=100)


@pytest.mark.parametrize("method,expected_status", REDIRECT_METHODS)
@pytest.mark.asyncio
async def test_redirect_method_bytes_location(method, expected_status, app):
    @app.router.get("/")
    async def home():
        return method(b"https://foo.org/somewhere")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == expected_status
    location = response.headers.get_single(b"location")
    assert location == b"https://foo.org/somewhere"


@pytest.mark.parametrize("method,expected_status", STATUS_METHODS_NO_BODY)
@pytest.mark.asyncio
async def test_no_body_method(method, expected_status, app):
    @app.router.get("/")
    async def home():
        return method()

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == expected_status
    assert response.has_body() is False


@pytest.mark.parametrize("method,expected_status", STATUS_METHODS_OPTIONAL_BODY)
@pytest.mark.asyncio
async def test_status_method_response(method, expected_status, app):
    @app.router.get("/")
    async def home():
        return method("Everything's good")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == expected_status
    content = await app.response.text()
    assert content == "Everything's good"


@pytest.mark.parametrize("method,expected_status", STATUS_METHODS_OPTIONAL_BODY)
@pytest.mark.asyncio
async def test_status_method_response_empty_body(method, expected_status, app):
    @app.router.get("/")
    async def home():
        return method()

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == expected_status
    content = await app.response.text()
    assert content == ""


@pytest.mark.parametrize("method,expected_status", STATUS_METHODS_OPTIONAL_BODY)
@pytest.mark.asyncio
async def test_status_method_response_with_object(method, expected_status, app):
    @app.router.get("/")
    async def home():
        return method(Foo(uuid4(), "foo", True))

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == expected_status
    content = await app.response.json()
    assert content.get("name") == "foo"
    assert content.get("ufo") is True


@pytest.mark.parametrize("status", [200, 201, 202, 400, 404, 500])
@pytest.mark.asyncio
async def test_status_code_response_with_text(status: int, app):
    @app.router.get("/")
    async def home():
        return status_code(status, "Everything's good")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == status
    content = await app.response.text()
    assert content == "Everything's good"


@pytest.mark.parametrize("status", [200, 201, 202, 400, 404, 500])
@pytest.mark.asyncio
async def test_status_code_response_with_empty_body(status: int, app):
    @app.router.get("/")
    async def home():
        return status_code(status)

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == status
    content = await app.response.text()
    assert content == ""


@pytest.mark.parametrize("status", [200, 201, 202, 400, 404, 500])
@pytest.mark.asyncio
async def test_status_code_response_with_object(status: int, app):
    @app.router.get("/")
    async def home():
        return status_code(status, Foo(uuid4(), "foo", True))

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == status
    content = await app.response.json()
    assert content.get("name") == "foo"
    assert content.get("ufo") is True


@pytest.mark.asyncio
async def test_created_response_with_empty_body(app):
    @app.router.get("/")
    async def home():
        return created(location="https://foo.org/foo/001")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 201
    location = response.headers.get_single(b"location")
    assert location == b"https://foo.org/foo/001"


@pytest.mark.asyncio
async def test_created_response_with_value(app):
    @app.router.get("/")
    async def home():
        return created(
            Foo(UUID("726807b3-5a82-4a59-8bed-65639d3529ba"), "example", False),
            location="https://foo.org/foo/726807b3-5a82-4a59-8bed-65639d3529ba",
        )

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 201
    location = response.headers.get_single(b"location")
    assert location == b"https://foo.org/foo/726807b3-5a82-4a59-8bed-65639d3529ba"
    content = await app.response.json()
    assert content.get("id") == "726807b3-5a82-4a59-8bed-65639d3529ba"
    assert content.get("name") == "example"
    assert content.get("ufo") is False


@pytest.mark.asyncio
async def test_text_response_default_status(app):
    @app.router.get("/")
    async def home():
        return text("Hello World")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"text/plain; charset=utf-8"

    body = await response.text()
    assert body == "Hello World"


@pytest.mark.parametrize(
    "content,status",
    [
        ("Hello World", 200),
        ("Hello World", 400),
        ("🍎 🐍", 200),
        ("""Lorem ipsum dolor sit amet""", 200),
    ],
)
@pytest.mark.asyncio
async def test_text_response_with_status(content, status, app):
    @app.router.get("/")
    async def home():
        return text(content, status)

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == status

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"text/plain; charset=utf-8"

    body = await response.text()
    assert body == content


@pytest.mark.asyncio
async def test_html_response_default_status(app):
    @app.router.get("/")
    async def home():
        return html(EXAMPLE_HTML)

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"text/html; charset=utf-8"

    body = await response.text()
    assert body == EXAMPLE_HTML


@pytest.mark.parametrize(
    "content,status",
    [
        (EXAMPLE_HTML, 200),
        (EXAMPLE_HTML, 400),
        ("<div><h2>Partial View</h2><p>Lorem ipsum dolor sit amet</p></div>", 200),
        ("<div><h2>Partial View</h2><p>Lorem ipsum dolor sit amet</p></div>", 400),
    ],
)
@pytest.mark.asyncio
async def test_html_response_with_status(content, status, app):
    @app.router.get("/")
    async def home():
        return html(content, status)

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == status

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"text/html; charset=utf-8"

    body = await response.text()
    assert body == content


@pytest.mark.parametrize("obj,values", JSON_OBJECTS)
@pytest.mark.asyncio
async def test_json_response(obj: Any, values: Dict[str, Any], app):
    @app.router.get("/")
    async def home():
        return json(obj)

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"application/json"

    raw = await response.text()
    data = await response.json()
    for name, value in values.items():
        assert data.get(name) == value

        if isinstance(value, str):
            assert f'"{name}":"{value}"' in raw
        else:
            assert f'"{name}":' in raw


@pytest.mark.parametrize("obj,values", JSON_OBJECTS)
@pytest.mark.asyncio
async def test_pretty_json_response(obj: Any, values: Dict[str, Any], app):
    @app.router.get("/")
    async def home():
        return pretty_json(obj)

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"application/json"

    raw = await response.text()
    data = await response.json()
    for name, value in values.items():
        assert data.get(name) == value
        if isinstance(value, str):
            assert f'    "{name}": "{value}"' in raw
        else:
            assert f'    "{name}": ' in raw


@pytest.mark.asyncio
async def test_file_response_from_fs(app):
    file_path = get_file_path("example.config", "files2")

    @app.router.get("/")
    async def home():
        return file(file_path, "text/plain; charset=utf-8")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/plain; charset=utf-8"
    assert response.headers.get_single(b"content-disposition") == b"attachment"

    text = await response.text()
    with open(file_path, mode="rt", encoding="utf8") as f:
        contents = f.read()
        assert contents == text


@pytest.mark.asyncio
async def test_file_response_from_fs_with_filename(app):
    file_path = get_file_path("example.config", "files2")

    @app.router.get("/")
    async def home():
        return file(file_path, "text/plain; charset=utf-8", file_name="foo.xml")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/plain; charset=utf-8"
    assert (
        response.headers.get_single(b"content-disposition")
        == b'attachment; filename="foo.xml"'
    )

    text = await response.text()
    with open(file_path, mode="rt", encoding="utf8") as f:
        contents = f.read()
        assert contents == text


async def get_example_css():
    yield b"body {\n"
    yield b"    background-color: red;\n"
    yield b"}\n"
    yield b"\n"
    yield b"h1 {\n"
    yield b"    font-size: 2rem;\n"
    yield b"}\n"


@pytest.mark.asyncio
async def test_file_response_from_generator(app):
    @app.router.get("/")
    async def home():
        return file(get_example_css, "text/css")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/css"
    assert response.headers.get_single(b"content-disposition") == b"attachment"

    text = await response.text()
    assert text == await read_from_asynciterable(get_example_css)


@pytest.mark.asyncio
async def test_file_response_from_bytes_io(app):
    bytes_io: Optional[BytesIO] = None

    @app.router.get("/")
    async def home():
        nonlocal bytes_io
        bytes_io = BytesIO()
        bytes_io.write("Żywią i bronią".encode("utf-8"))
        return file(bytes_io, "text/plain; charset=utf-8", file_name="foo.txt")

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/plain; charset=utf-8"
    assert (
        response.headers.get_single(b"content-disposition")
        == b'attachment; filename="foo.txt"'
    )

    assert bytes_io is not None
    assert bytes_io.closed  # type: ignore


@pytest.mark.asyncio
async def test_file_response_from_generator_inline(app):
    @app.router.get("/")
    async def home():
        return file(
            get_example_css,
            "text/css",
            content_disposition=ContentDispositionType.INLINE,
        )

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/css"
    assert response.headers.get_single(b"content-disposition") == b"inline"

    text = await response.text()
    assert text == await read_from_asynciterable(get_example_css)


@pytest.mark.asyncio
async def test_file_response_from_bytes(app):
    @app.router.get("/")
    async def home():
        return file(
            EXAMPLE_HTML.encode("utf8"),
            "text/css",
            file_name="home.css",
            content_disposition=ContentDispositionType.INLINE,
        )

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/css"
    assert (
        response.headers.get_single(b"content-disposition")
        == b'inline; filename="home.css"'
    )

    text = await response.text()
    assert text == EXAMPLE_HTML


@pytest.mark.asyncio
async def test_file_response_from_byte_array(app):
    value = bytearray()
    value.extend(b"Hello!\n")
    value.extend(b"World!\n\n")
    value.extend(b"...")
    expected_result = bytes(value).decode("utf8")

    @app.router.get("/")
    async def home():
        return file(
            value,
            "text/plain",
            file_name="example.txt",
            content_disposition=ContentDispositionType.INLINE,
        )

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/plain"
    assert (
        response.headers.get_single(b"content-disposition")
        == b'inline; filename="example.txt"'
    )

    text = await response.text()
    assert text == expected_result


@pytest.mark.asyncio
async def test_file_response_from_generator_inline_with_name(app):
    @app.router.get("/")
    async def home():
        return file(
            get_example_css,
            "text/css",
            file_name="home.css",
            content_disposition=ContentDispositionType.INLINE,
        )

    app.normalize_handlers()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200
    assert response.headers.get_single(b"content-type") == b"text/css"
    assert (
        response.headers.get_single(b"content-disposition")
        == b'inline; filename="home.css"'
    )

    text = await response.text()
    assert text == await read_from_asynciterable(get_example_css)


def test_files_raises_for_invalid_input():
    with pytest.raises(ValueError):
        file(True, "text/plain")  # type: ignore

    with pytest.raises(ValueError):
        file(100, "text/plain")  # type: ignore

    with pytest.raises(ValueError):
        file([10, 120, 400], "text/plain")  # type: ignore


def test_files_raises_for_invalid_name_with_folder_path():
    with pytest.raises(ValueError):
        file(b"Hello, There!", "text/plain", file_name="not_good/")


def test_json_response_raises_for_not_json_serializable():
    class NotSerializable:
        def __init__(self) -> None:
            self.foo = True
            self.might_be_secret = "iunisud109283012"

    with pytest.raises(TypeError):
        json(NotSerializable())


@pytest.mark.parametrize("status", [200, 201, 202, 400, 404, 500])
@pytest.mark.asyncio
async def test_status_code_response_with_text_in_controller(status: int, app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        def greet(self):
            return self.status_code(status, "Everything's good")

    await app.start()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == status
    content = await app.response.text()
    assert content == "Everything's good"


@pytest.mark.parametrize("method,expected_status", STATUS_METHODS_OPTIONAL_BODY)
@pytest.mark.asyncio
async def test_status_method_response_in_controller(method, expected_status, app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    method_name = method.__name__

    class Home(Controller):
        @get("/")
        def greet(self):
            controller_method = getattr(Home, method_name, None)
            return controller_method(self, "Everything's good")

    controller_method = getattr(Home, method_name, None)
    assert callable(controller_method)

    await app.start()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == expected_status
    content = await app.response.text()
    assert content == "Everything's good"


@pytest.mark.parametrize("method,expected_status", STATUS_METHODS_NO_BODY)
@pytest.mark.asyncio
async def test_status_method_without_body_response_in_controller(
    method, expected_status, app
):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    method_name = method.__name__

    class Home(Controller):
        @get("/")
        def greet(self):
            controller_method = getattr(Home, method_name, None)
            return controller_method(self)

    controller_method = getattr(Home, method_name, None)
    assert callable(controller_method)

    await app.start()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == expected_status
    assert response.has_body() is False


@pytest.mark.parametrize("method,expected_status", REDIRECT_METHODS)
@pytest.mark.asyncio
async def test_redirect_method_in_controller(method, expected_status, app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    method_name = method.__name__

    class Home(Controller):
        @get("/")
        def greet(self):
            controller_method = getattr(Home, method_name, None)
            return controller_method(self, "https://foo.org/somewhere")

    controller_method = getattr(Home, method_name, None)
    assert callable(controller_method)

    await app.start()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == expected_status
    location = response.headers.get_single(b"location")
    assert location == b"https://foo.org/somewhere"


@pytest.mark.asyncio
async def test_view_methods_in_controller_throw_if_templating_not_configured(app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        def greet(self):
            self.view("foo")

    await app.start()

    home = Home()
    with pytest.raises(TemplatingNotConfiguredException):
        home.greet()

    with pytest.raises(TemplatingNotConfiguredException):
        await home.view_async("foo")


@pytest.mark.asyncio
async def test_view_methods_in_controller_throw_if_view_name_cannot_be_determined():
    class Home(Controller):
        pass

    home = Home()
    with pytest.raises(CannotDetermineDefaultViewNameError):
        home.view()

    with pytest.raises(CannotDetermineDefaultViewNameError):
        await home.view_async()


@pytest.mark.asyncio
async def test_view_methods_in_controller_throw_if_sys_get_frame_is_not_defined():
    class Home(Controller):
        pass

    def monkey_getframe(num: int):
        raise AttributeError("Test")

    base_getframe = sys._getframe
    sys._getframe = monkey_getframe

    home = Home()
    with pytest.raises(CannotDetermineDefaultViewNameError):
        home.view()

    sys._getframe = base_getframe


@pytest.mark.parametrize("obj,values", JSON_OBJECTS)
@pytest.mark.asyncio
async def test_json_response_in_controller(obj: Any, values: Dict[str, Any], app):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        def greet(self):
            return self.json(obj)

    await app.start()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    assert response.status == 200

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"application/json"

    raw = await response.text()
    data = await response.json()
    for name, value in values.items():
        assert data.get(name) == value

        if isinstance(value, str):
            assert f'"{name}":"{value}"' in raw
        else:
            assert f'"{name}":' in raw


@pytest.mark.parametrize("obj,values", JSON_OBJECTS)
@pytest.mark.asyncio
async def test_pretty_json_response_in_controller(
    obj: Any, values: Dict[str, Any], app
):
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Home(Controller):
        @get("/")
        def greet(self):
            return self.pretty_json(obj)

    await app.start()

    await app(
        get_example_scope("GET", "/", []),
        MockReceive(),
        MockSend(),
    )

    response = app.response
    raw = await response.text()
    assert response.status == 200

    content_type = response.headers.get_single(b"content-type")
    assert content_type == b"application/json"

    raw = await response.text()
    data = await response.json()
    for name, value in values.items():
        assert data.get(name) == value
        if isinstance(value, str):
            assert f'    "{name}": "{value}"' in raw
        else:
            assert f'    "{name}": ' in raw



import json
from typing import Any

from essentials.json import dumps


def default_json_dumps(obj):
    return dumps(obj, separators=(",", ":"))


def default_pretty_json_dumps(obj):
    return dumps(obj, indent=4)


class JSONPlugin:
    def __init__(self):
        self._loads = json.loads
        self._dumps = default_json_dumps
        self._pretty_dumps = default_pretty_json_dumps

    def use(
        self,
        loads=json.loads,
        dumps=default_json_dumps,
        pretty_dumps=default_pretty_json_dumps,
    ):
        self._loads = loads
        self._dumps = dumps
        self._pretty_dumps = pretty_dumps

    def loads(self, text: str) -> Any:
        return self._loads(text)

    def dumps(self, obj: Any) -> str:
        return self._dumps(obj)

    def pretty_dumps(self, obj: Any) -> str:
        return self._pretty_dumps(obj)



import pytest

from blacksheep.common.files.pathsutils import (
    get_file_extension_from_name,
    get_mime_type_from_name,
)


@pytest.mark.parametrize(
    "full_path,expected_result",
    [
        ("hello.txt", ".txt"),
        (".gitignore", ".gitignore"),
        ("ØØ Void.album", ".album"),
        ("", ""),
    ],
)
def test_get_file_extension_from_name(full_path, expected_result):
    assert get_file_extension_from_name(full_path) == expected_result


@pytest.mark.parametrize(
    "full_path,expected_result",
    [
        ("example.ogg", "audio/ogg"),
        ("example.jpg", "image/jpeg"),
        ("example.jpeg", "image/jpeg"),
        ("example.png", "image/png"),
        ("example.js", "application/javascript"),
        ("example.json", "application/json"),
        ("example.woff2", "font/woff2"),
        ("hello.txt", "text/plain"),
        (".gitignore", "application/octet-stream"),
        ("ØØ Void.album", "application/octet-stream"),
        ("", "application/octet-stream"),
    ],
)
def test_get_mime_type(full_path, expected_result):
    assert get_mime_type_from_name(full_path) == expected_result



import pytest

from blacksheep import Content, Request, scribe
from blacksheep.contents import FormPart, MultiPartFormData
from blacksheep.exceptions import BadRequestFormat
from blacksheep.messages import get_absolute_url_to_path, get_request_absolute_url
from blacksheep.scribe import write_request, write_small_request
from blacksheep.server.asgi import (
    get_request_url,
    get_request_url_from_scope,
    incoming_request,
)
from blacksheep.testing.helpers import get_example_scope
from blacksheep.url import URL


def test_request_supports_dynamic_attributes():
    request = Request("GET", b"/", None)
    foo = object()

    assert (
        hasattr(request, "foo") is False
    ), "This test makes sense if such attribute is not defined"
    request.foo = foo  # type: ignore
    assert request.foo is foo  # type: ignore


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "url,method,headers,content,expected_result",
    [
        (
            b"https://robertoprevato.github.io",
            "GET",
            [],
            None,
            b"GET / HTTP/1.1\r\nhost: robertoprevato.github.io\r\ncontent-length: 0\r\n\r\n",
        ),
        (
            b"https://robertoprevato.github.io",
            "HEAD",
            [],
            None,
            b"HEAD / HTTP/1.1\r\nhost: robertoprevato.github.io\r\ncontent-length: 0\r\n\r\n",
        ),
        (
            b"https://robertoprevato.github.io",
            "POST",
            [],
            None,
            b"POST / HTTP/1.1\r\nhost: robertoprevato.github.io\r\ncontent-length: 0\r\n\r\n",
        ),
        (
            b"https://robertoprevato.github.io/How-I-created-my-own-media-storage-in-Azure/",
            "GET",
            [],
            None,
            b"GET /How-I-created-my-own-media-storage-in-Azure/ HTTP/1.1\r\nhost: robertoprevato.github.io"
            b"\r\ncontent-length: 0\r\n\r\n",
        ),
        (
            b"https://foo.org/a/b/c/?foo=1&ufo=0",
            "GET",
            [],
            None,
            b"GET /a/b/c/?foo=1&ufo=0 HTTP/1.1\r\nhost: foo.org\r\ncontent-length: 0\r\n\r\n",
        ),
    ],
)
async def test_request_writing(url, method, headers, content, expected_result):
    request = Request(method, url, headers).with_content(content)
    data = b""
    async for chunk in scribe.write_request(request):
        data += chunk
    assert data == expected_result


@pytest.mark.parametrize(
    "url,query,parsed_query",
    [
        (b"https://foo.org/a/b/c?hello=world", b"hello=world", {"hello": ["world"]}),
        (
            b"https://foo.org/a/b/c?hello=world&foo=power",
            b"hello=world&foo=power",
            {"hello": ["world"], "foo": ["power"]},
        ),
        (
            b"https://foo.org/a/b/c?hello=world&foo=power&foo=200",
            b"hello=world&foo=power&foo=200",
            {"hello": ["world"], "foo": ["power", "200"]},
        ),
    ],
)
def test_parse_query(url, query, parsed_query):
    request = Request("GET", url, None)
    assert request.url.value == url
    assert request.url.query == query
    assert request.query == parsed_query


@pytest.mark.asyncio
async def test_can_read_json_data_even_without_content_type_header():
    request = Request("POST", b"/", None)

    request.with_content(Content(b"application/json", b'{"hello":"world","foo":false}'))

    json = await request.json()
    assert json == {"hello": "world", "foo": False}


@pytest.mark.asyncio
async def test_if_read_json_fails_content_type_header_is_checked_json_gives_bad_request_format():
    request = Request("POST", b"/", [(b"Content-Type", b"application/json")])

    request.with_content(Content(b"application/json", b'{"hello":'))  # broken json

    with pytest.raises(BadRequestFormat):
        await request.json()


@pytest.mark.asyncio
async def test_if_read_json_fails_content_type_header_is_checked_non_json_gives_invalid_operation():
    request = Request("POST", b"/", [])

    request.with_content(
        Content(b"application/json", b'{"hello":')
    )  # broken json; broken content-type

    with pytest.raises(BadRequestFormat):
        await request.json()


def test_cookie_parsing():
    request = Request(
        "POST", b"/", [(b"Cookie", b"ai=something; hello=world; foo=Hello%20World%3B;")]
    )

    assert request.cookies == {
        "ai": "something",
        "hello": "world",
        "foo": "Hello World;",
    }


def test_cookie_parsing_multiple_cookie_headers():
    request = Request(
        "POST",
        b"/",
        [
            (b"Cookie", b"ai=something; hello=world; foo=Hello%20World%3B;"),
            (b"Cookie", b"jib=jab; ai=else;"),
        ],
    )

    assert request.cookies == {
        "ai": "else",
        "hello": "world",
        "foo": "Hello World;",
        "jib": "jab",
    }


def test_cookie_parsing_duplicated_cookie_header_value():
    request = Request(
        "POST",
        b"/",
        [(b"Cookie", b"ai=something; hello=world; foo=Hello%20World%3B; hello=kitty;")],
    )

    assert request.cookies == {
        "ai": "something",
        "hello": "kitty",
        "foo": "Hello World;",
    }


@pytest.mark.parametrize(
    "header,expected_result",
    [
        [(b"Expect", b"100-Continue"), True],
        [(b"expect", b"100-continue"), True],
        [(b"X-Foo", b"foo"), False],
    ],
)
def test_request_expect_100_continue(header, expected_result):
    request = Request("POST", b"/", [header])
    assert expected_result == request.expect_100_continue()


@pytest.mark.parametrize(
    "headers,expected_result",
    [
        [[(b"Content-Type", b"application/json")], True],
        [[(b"Content-Type", b"application/problem+json")], True],
        [[(b"Content-Type", b"application/json; charset=utf-8")], True],
        [[], False],
        [[(b"Content-Type", b"application/xml")], False],
    ],
)
def test_request_declares_json(headers, expected_result):
    request = Request("GET", b"/", headers)
    assert request.declares_json() is expected_result


def test_small_request_headers_add_through_higher_api():
    request = Request("GET", b"https://hello-world", None)

    request.headers.add(b"Hello", b"World")

    raw_bytes = write_small_request(request)

    assert b"Hello: World\r\n" in raw_bytes


def test_small_request_headers_add_through_higher_api_many():
    request = Request("GET", b"https://hello-world", None)

    request.headers.add_many({b"Hello": b"World", b"X-Foo": b"Foo"})

    raw_bytes = write_small_request(request)

    assert b"Hello: World\r\n" in raw_bytes
    assert b"X-Foo: Foo\r\n" in raw_bytes


def test_small_request_headers_add_through_lower_api():
    request = Request("GET", b"https://hello-world", None)

    request.add_header(b"Hello", b"World")

    raw_bytes = write_small_request(request)

    assert b"Hello: World\r\n" in raw_bytes


@pytest.mark.parametrize(
    "initial_url,new_url",
    [
        (b"https://hello-world/", b"https://ciao-mondo/"),
        (b"https://hello-world/one/two/three", b"https://hello-world/one/two/three/"),
        (b"https://hello-world/one/two/three/", b"https://hello-world/one/two/three"),
    ],
)
def test_request_can_update_url(initial_url, new_url):
    request = Request("GET", initial_url, None)

    assert request.url.value == initial_url

    request.url = URL(new_url)

    assert request.url.value == new_url


def test_request_content_type_is_read_from_content():
    request = Request("POST", b"/", []).with_content(
        MultiPartFormData([FormPart(b"a", b"world"), FormPart(b"b", b"9000")])
    )

    assert request.content is not None
    assert request.content_type() == request.content.type


@pytest.mark.parametrize(
    "scope,expected_value",
    [
        (
            get_example_scope("GET", "/foo", scheme="http", server=["127.0.0.1", 8000]),
            "http://127.0.0.1:8000/foo",
        ),
        (
            get_example_scope("GET", "/foo", scheme="http", server=["127.0.0.1", 80]),
            "http://127.0.0.1/foo",
        ),
        (
            get_example_scope(
                "GET", "/foo", scheme="https", server=["127.0.0.1", 44777]
            ),
            "https://127.0.0.1:44777/foo",
        ),
        (
            get_example_scope("GET", "/foo", scheme="https", server=["127.0.0.1", 443]),
            "https://127.0.0.1/foo",
        ),
    ],
)
def test_get_asgi_request_full_url(scope, expected_value):
    request = incoming_request(scope, None)

    full_url = get_request_url(request)
    assert full_url == expected_value


def test_request_pyi():
    request = Request("GET", b"/", [(b"cookie", b"foo=aaa")])

    request.cookies["foo"] == "aaa"
    request.get_cookie("foo") == "aaa"
    request.get_first_header(b"cookie") == b"foo=aaa"

    request.set_cookie("lorem", "ipsum")
    request.get_cookie("lorem") == "ipsum"


@pytest.mark.parametrize(
    "scope,trailing_slash,expected_value",
    [
        [
            {"scheme": "https", "path": "/", "server": ("www.neoteroi.dev", 443)},
            False,
            "https://www.neoteroi.dev/",
        ],
        [
            {"scheme": "https", "path": "/admin", "server": ("www.neoteroi.dev", 443)},
            False,
            "https://www.neoteroi.dev/admin",
        ],
        [
            {"scheme": "https", "path": "/admin", "server": ("www.neoteroi.dev", 443)},
            True,
            "https://www.neoteroi.dev/admin/",
        ],
        [
            {
                "scheme": "https",
                "path": "/admin",
                "server": ("www.neoteroi.dev", 44777),
            },
            True,
            "https://www.neoteroi.dev:44777/admin/",
        ],
        [
            {"scheme": "http", "path": "/admin", "server": ("www.neoteroi.dev", 44777)},
            True,
            "http://www.neoteroi.dev:44777/admin/",
        ],
        [
            {"scheme": "http", "path": "/admin", "server": ("www.neoteroi.dev", 80)},
            True,
            "http://www.neoteroi.dev/admin/",
        ],
        [
            {
                "scheme": "http",
                "path": "/admin",
                "server": ("www.neoteroi.dev", 80),
                "query_string": b"foo=Hello%20World%20%C3%B8",
            },
            False,
            "http://www.neoteroi.dev/admin?foo=Hello%20World%20%C3%B8",
        ],
    ],
)
def test_get_request_url_from_scope(scope, trailing_slash, expected_value):
    result = get_request_url_from_scope(scope, trailing_slash=trailing_slash)
    assert result == expected_value


def test_get_request_url_from_scope_raises_for_invalid_scope():
    with pytest.raises(ValueError):
        get_request_url_from_scope({})


@pytest.mark.parametrize(
    "scope,expected_value",
    [
        (
            get_example_scope("GET", "/foo", scheme="http", server=["127.0.0.1", 8000]),
            "http://127.0.0.1:8000/foo",
        ),
        (
            get_example_scope("GET", "/foo", scheme="http", server=["127.0.0.1", 80]),
            "http://127.0.0.1/foo",
        ),
        (
            get_example_scope(
                "GET", "/foo", scheme="https", server=["127.0.0.1", 44777]
            ),
            "https://127.0.0.1:44777/foo",
        ),
        (
            get_example_scope("GET", "/foo", scheme="https", server=["127.0.0.1", 443]),
            "https://127.0.0.1/foo",
        ),
    ],
)
def test_get_request_absolute_url(scope, expected_value):
    request = incoming_request(scope)

    assert request.scheme == scope["scheme"]
    assert request.host == dict(scope["headers"])[b"host"].decode()
    assert request.base_path == ""

    absolute_url = get_request_absolute_url(request)
    assert str(absolute_url) == f"{request.scheme}://{request.host}{request.path}"
    assert str(absolute_url) == expected_value


@pytest.mark.parametrize(
    "scope,base_path,expected_value",
    [
        (
            get_example_scope("GET", "/foo", scheme="http", server=["127.0.0.1", 8000]),
            "/api",
            "http://127.0.0.1:8000/api/foo",
        ),
        (
            get_example_scope("GET", "/foo", scheme="http", server=["127.0.0.1", 80]),
            "/api/",
            "http://127.0.0.1/api/foo",
        ),
        (
            get_example_scope(
                "GET", "/foo", scheme="https", server=["127.0.0.1", 44777]
            ),
            "/api/oof",
            "https://127.0.0.1:44777/api/oof/foo",
        ),
        (
            get_example_scope("GET", "/foo", scheme="https", server=["127.0.0.1", 443]),
            "/api/oof/",
            "https://127.0.0.1/api/oof/foo",
        ),
    ],
)
def test_get_request_absolute_url_with_base_path(scope, base_path, expected_value):
    request = incoming_request(scope)

    assert request.scheme == scope["scheme"]
    assert request.host == dict(scope["headers"])[b"host"].decode()
    request.base_path = base_path

    absolute_url = get_request_absolute_url(request)
    assert str(absolute_url) == expected_value


@pytest.mark.parametrize(
    "scope,path,expected_result",
    [
        (
            get_example_scope("GET", "/foo", scheme="http", server=["127.0.0.1", 8000]),
            "/sign-in",
            "http://127.0.0.1:8000/sign-in",
        ),
        (
            get_example_scope("GET", "/", scheme="http", server=["127.0.0.1", 8000]),
            "/authorization/callback",
            "http://127.0.0.1:8000/authorization/callback",
        ),
        (
            get_example_scope(
                "GET", "/a/b/c/", scheme="http", server=["127.0.0.1", 8000]
            ),
            "/authorization/callback",
            "http://127.0.0.1:8000/authorization/callback",
        ),
    ],
)
def test_get_request_absolute_url_to_path(scope, path, expected_result):
    request = incoming_request(scope)
    url_to = get_absolute_url_to_path(request, path)

    assert str(url_to) == expected_result


def test_can_set_request_host_and_scheme():
    scope = get_example_scope(
        "GET", "/blacksheep/", scheme="http", server=["127.0.0.1", 80]
    )
    request = incoming_request(scope)

    request.scheme = "https"
    request.host = "neoteroi.dev"

    absolute_url = get_request_absolute_url(request)
    assert str(absolute_url) == "https://neoteroi.dev/blacksheep/"


def test_can_set_request_client_ip():
    scope = get_example_scope(
        "GET", "/blacksheep/", scheme="http", server=["127.0.0.1", 80]
    )
    request = incoming_request(scope)

    request.client_ip == scope["client"][0]

    assert request.original_client_ip == "127.0.0.1"

    # can set (e.g. when handling forwarded headers)
    request.original_client_ip = "185.152.122.103"

    assert request.original_client_ip == "185.152.122.103"
    assert scope["client"] == ("127.0.0.1", 51492)


def test_updating_request_url_read_host():
    request = Request("GET", b"https://www.neoteroi.dev/blacksheep", [])

    assert request.path == "/blacksheep"
    assert request.host == "www.neoteroi.dev"

    request.url = "https://github.com/RobertoPrevato"

    assert request.path == "/RobertoPrevato"
    assert request.host == "github.com"


@pytest.mark.asyncio
async def test_updating_request_host_in_headers():
    request = Request("GET", b"https://www.neoteroi.dev/blacksheep", [])

    assert request.path == "/blacksheep"
    assert request.host == "www.neoteroi.dev"
    assert request.headers[b"host"] == tuple()

    async for _ in write_request(request):
        pass

    assert request.headers[b"host"] == (b"www.neoteroi.dev",)
    request.url = "https://github.com/RobertoPrevato"

    assert request.headers[b"host"] == tuple()

    async for _ in write_request(request):
        pass

    assert request.headers[b"host"] == (b"github.com",)


@pytest.mark.asyncio
async def test_write_request_cookies():
    request = Request("GET", b"https://www.neoteroi.dev/blacksheep", [])

    request.set_cookie("example_1", "one")
    request.set_cookie("example_2", "two")

    expected_bytes = b"""GET /blacksheep HTTP/1.1
cookie: example_1=one;example_2=two
host: www.neoteroi.dev
content-length: 0

""".replace(
        b"\n", b"\r\n"
    )

    value = bytearray()

    async for chunk in write_request(request):
        value.extend(chunk)

    assert bytes(value) == expected_bytes




cdef class InvalidOperation(Exception):

    def __init__(self, str message, object inner_exception=None):
        super().__init__(message)
        self.inner_exception = inner_exception


cdef class HTTPException(Exception):

    def __init__(self, int status, str message = "HTTP exception"):
        super().__init__(message)
        self.status = status


cdef class BadRequest(HTTPException):

    def __init__(self, message=None):
        super().__init__(400, message or "Bad request")


cdef class BadRequestFormat(BadRequest):

    def __init__(self, str message, object inner_exception=None):
        super().__init__(message)
        self.inner_exception = inner_exception


cdef class NotFound(HTTPException):

    def __init__(self, message=None):
        super().__init__(404, message or "Not found")


cdef class Unauthorized(HTTPException):

    def __init__(self, message=None):
        super().__init__(401, message or "Unauthorized")


cdef class Forbidden(HTTPException):

    def __init__(self, message=None):
        super().__init__(403, message or "Forbidden")


cdef class RangeNotSatisfiable(HTTPException):

    def __init__(self):
        super().__init__(416, "Range not satisfiable")


cdef class InternalServerError(HTTPException):

    def __init__(self):
        super().__init__(500, "Internal server error")


cdef class NotImplementedByServer(HTTPException):

    def __init__(self):
        super().__init__(501, "Not implemented by server")


cdef class InvalidArgument(Exception):

    def __init__(self, str message):
        super().__init__(message)


cdef class MessageAborted(Exception):
    def __init__(self):
        super().__init__(
            "The message was aborted before the client sent its whole content."
        )



from typing import AnyStr, Sequence

import pytest

from blacksheep.utils import ensure_bytes, ensure_str, join_fragments


@pytest.mark.parametrize(
    "fragments,expected_value",
    [
        [["a"], "/a"],
        [["a", "b", "c", "d"], "/a/b/c/d"],
        [["a", None, "b", "c", "", "d"], "/a/b/c/d"],
        [[b"a", b"b", b"c", b"d"], "/a/b/c/d"],
        [[b"a", "b", "c", b"d"], "/a/b/c/d"],
        [["hello/world", "today"], "/hello/world/today"],
        [[b"hello/world", b"today"], "/hello/world/today"],
        [["//hello///world", "/today/"], "/hello/world/today"],
    ],
)
def test_join_url_fragments(fragments: Sequence[AnyStr], expected_value: str):
    joined = join_fragments(*fragments)
    assert joined == expected_value


@pytest.mark.parametrize(
    "value,expected_result", [("hello", b"hello"), (b"hello", b"hello")]
)
def test_ensure_bytes(value, expected_result):
    assert ensure_bytes(value) == expected_result


@pytest.mark.parametrize(
    "value,expected_result", [("hello", "hello"), (b"hello", "hello")]
)
def test_ensure_str(value, expected_result):
    assert ensure_str(value) == expected_result


def test_ensure_bytes_throws_for_invalid_value():
    with pytest.raises(ValueError):
        ensure_bytes(True)  # type: ignore


def test_ensure_str_throws_for_invalid_value():
    with pytest.raises(ValueError):
        ensure_str(True)  # type: ignore



from ipaddress import ip_address, ip_network

import pytest

from blacksheep.server.remotes.forwarding import (
    ForwardedHeaderEntry,
    ForwardedHeadersMiddleware,
    XForwardedHeadersMiddleware,
    parse_forwarded_header,
)
from blacksheep.server.remotes.hosts import TrustedHostsMiddleware
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


@pytest.mark.parametrize(
    "forwarded_host,forwarded_ip,forwarded_proto",
    [
        (b"neoteroi.dev", b"203.0.113.195", b"https"),
        (b"id42.example-cdn.com", b"2001:db8:85a3:8d3:1319:8a2e:370:7348", b"http"),
    ],
)
@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware(
    app: FakeApplication,
    forwarded_host,
    forwarded_ip,
    forwarded_proto,
):
    app.middlewares.append(XForwardedHeadersMiddleware())

    @app.router.get("/")
    async def home(request):
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", forwarded_host),
            (b"X-Forwarded-For", forwarded_ip),
            (b"X-Forwarded-Proto", forwarded_proto),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != forwarded_host

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    last_request = app.request
    assert last_request is not None
    assert app.response is not None
    assert app.response.status == 204

    # the request is updated to reflect X-Forwarded information
    assert last_request.host == forwarded_host.decode()
    assert last_request.scheme == forwarded_proto.decode()
    assert last_request.original_client_ip == forwarded_ip.decode()


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_multiple_proxies(app: FakeApplication):
    app.middlewares.append(
        XForwardedHeadersMiddleware(
            forward_limit=3,
            known_proxies=[
                ip_address("127.0.0.1"),
                ip_address("203.0.113.196"),
                ip_address("203.0.113.197"),
            ],
        )
    )

    @app.router.get("/")
    async def home(request):
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    last_request = app.request
    assert last_request is not None
    assert app.response is not None
    assert app.response.status == 204

    # the request is updated to reflect X-Forwarded information
    assert last_request.host == "neoteroi.dev"
    assert last_request.scheme == "https"
    assert last_request.original_client_ip == "203.0.113.195"

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195,203.0.113.196,203.0.113.197"),
            (b"X-Forwarded-Proto", b"https,http,http"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    await app(scope, MockReceive(), MockSend())

    last_request = app.request
    assert last_request is not None
    assert app.response is not None
    assert app.response.status == 204

    # the request is updated to reflect X-Forwarded information
    assert last_request.host == "neoteroi.dev"
    assert last_request.scheme == "https"
    assert last_request.original_client_ip == "203.0.113.195"


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_invalid_host(app: FakeApplication):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"ugly-domain.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400

    assert not called

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204

    assert called


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_without_forwarded_for(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_without_forwarded_proto(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_too_many_forwards(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195,203.0.113.196"),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_multiple_forwarded_host_headers(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_multiple_forwarded_proto_headers(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
            (b"X-Forwarded-Proto", b"http"),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_multiple_forwarded_proto_values(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https,http"),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_multiple_forwarded_for_headers(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-For", b"203.0.113.195,203.0.113.196"),
            (b"X-Forwarded-For", b"203.0.113.195"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_too_many_forward_values(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-For", b"203.0.113.195,203.0.113.196"),
            (b"X-Forwarded-Proto", b"https"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_invalid_host_not_forwarded(
    app: FakeApplication,
):
    app.middlewares.append(XForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_invalid_proxy_id(
    app: FakeApplication,
):
    app.middlewares.append(
        XForwardedHeadersMiddleware(known_proxies=[ip_address("185.152.122.103")])
    )

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400

    assert not called

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
        client=("185.152.122.103", 443),
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204

    assert called


@pytest.mark.asyncio
async def test_x_forwarded_headers_middleware_blocks_invalid_proxy_id_by_network(
    app: FakeApplication,
):
    app.middlewares.append(
        XForwardedHeadersMiddleware(known_networks=[ip_network("192.168.0.0/24")])
    )

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
        client=("203.0.113.196", 443),
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400

    assert not called

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (b"X-Forwarded-Host", b"neoteroi.dev"),
            (b"X-Forwarded-For", b"203.0.113.195"),
            (b"X-Forwarded-Proto", b"https"),
        ],
        client=("192.168.0.1", 443),
    )

    assert scope["scheme"] == "http"
    assert dict(scope["headers"])[b"host"] != b"neoteroi.dev"

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.request is not None
    assert app.response is not None
    assert app.response.status == 204
    assert app.request.original_client_ip == "203.0.113.195"

    assert called


@pytest.mark.asyncio
async def test_forwarded_header_middleware(app: FakeApplication):
    app.middlewares.append(ForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (
                b"Forwarded",
                b"for=_hidden;host=neoteroi.dev;proto=https",
            ),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204

    assert app.request is not None
    assert app.request.host == "neoteroi.dev"
    assert app.request.original_client_ip == "_hidden"
    assert app.request.scheme == "https"


@pytest.mark.asyncio
async def test_forwarded_header_middleware_by_invalid(app: FakeApplication):
    app.middlewares.append(ForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (
                b"Forwarded",
                b"for=_hidden;host=neoteroi.dev;proto=https;by=203.0.113.195",
            ),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


@pytest.mark.asyncio
async def test_forwarded_header_middleware_by_hidden(app: FakeApplication):
    app.middlewares.append(ForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (
                b"Forwarded",
                b"for=_hidden;host=neoteroi.dev;proto=https;by=_hidden",
            ),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204

    assert app.request is not None
    assert app.request.host == "neoteroi.dev"
    assert app.request.original_client_ip == "_hidden"
    assert app.request.scheme == "https"


@pytest.mark.asyncio
async def test_forwarded_header_middleware_blocks_requests_with_too_many_forwards(
    app: FakeApplication,
):
    app.middlewares.append(ForwardedHeadersMiddleware(allowed_hosts=["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope(
        "GET",
        "/",
        extra_headers=[
            (
                b"Forwarded",
                b"for=_hidden,for=_hidden",
            ),
        ],
    )

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400


def test_forwarded_entry_equality():
    a = ForwardedHeaderEntry(
        forwarded_for="203.0.113.195",
        forwarded_by="_proxy",
        forwarded_proto="https",
        forwarded_host="neoteroi.dev",
    )

    assert a == {
        "forwarded_for": "203.0.113.195",
        "forwarded_by": "_proxy",
        "forwarded_proto": "https",
        "forwarded_host": "neoteroi.dev",
    }

    assert a == ForwardedHeaderEntry(
        forwarded_for="203.0.113.195",
        forwarded_by="_proxy",
        forwarded_proto="https",
        forwarded_host="neoteroi.dev",
    )

    assert a != ForwardedHeaderEntry(
        forwarded_for="_hidden",
        forwarded_by="_proxy",
        forwarded_proto="https",
        forwarded_host="neoteroi.dev",
    )

    assert a != 2


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (
            "host=neoteroi.dev;for=203.0.113.195;proto=https;by=_secret",
            ForwardedHeaderEntry(
                forwarded_for="203.0.113.195",
                forwarded_by="_secret",
                forwarded_proto="https",
                forwarded_host="neoteroi.dev",
            ),
        ),
        (
            "host=neoteroi.dev",
            ForwardedHeaderEntry(
                forwarded_for="",
                forwarded_by="",
                forwarded_proto="",
                forwarded_host="neoteroi.dev",
            ),
        ),
        (
            "host=neoteroi.dev;proto=https",
            ForwardedHeaderEntry(
                forwarded_for="",
                forwarded_by="",
                forwarded_proto="https",
                forwarded_host="neoteroi.dev",
            ),
        ),
    ],
)
def test_parse_forwarded_entry(value, expected_result):
    parsed = next(iter(parse_forwarded_header(value)), None)
    assert parsed == expected_result


@pytest.mark.asyncio
async def test_trusted_hosts_middleware_blocks_invalid_host(app: FakeApplication):
    app.middlewares.append(TrustedHostsMiddleware(["neoteroi.dev"]))

    called = False

    @app.router.get("/")
    async def home(request):
        nonlocal called
        called = True
        return

    scope = get_example_scope("GET", "/", server=("ugly-domain.dev", 80))

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 400

    assert not called

    scope = get_example_scope("GET", "/", server=("neoteroi.dev", 80))

    app.prepare()
    await app(scope, MockReceive(), MockSend())

    assert app.response is not None
    assert app.response.status == 204

    assert called



"""
This module provides functions to configure Cache-Control headers for responses and
request handlers, including a decorator and a middleware.
"""

import inspect
from functools import wraps
from typing import Optional

from blacksheep import Request, Response
from blacksheep.normalization import copy_special_attributes
from blacksheep.server.normalization import ensure_response


def write_cache_control_response_header(
    *,
    max_age: Optional[int] = None,
    shared_max_age: Optional[int] = None,
    no_cache: Optional[bool] = None,
    no_store: Optional[bool] = None,
    must_revalidate: Optional[bool] = None,
    proxy_revalidate: Optional[bool] = None,
    private: Optional[bool] = None,
    public: Optional[bool] = None,
    must_understand: Optional[bool] = None,
    no_transform: Optional[bool] = None,
    immutable: Optional[bool] = None,
    stale_while_revalidate: Optional[int] = None,
    stale_if_error: Optional[int] = None,
) -> bytes:
    """
    Writes the value of a Cache-Control response header, applying the given directives.

    Parameters
    ----------
    max_age: int | None
        max age in seconds
    shared_max_age: int | None
        optional shared max age in seconds
    no_cache: bool | None
        enables no-cache
    no_store: bool | None
        enables no-store
    must_revalidate: bool | None
        enables must-revalidate
    proxy_revalidate: bool | None
        enables proxy-revalidate
    private: bool | None
        enables private
    public: bool | None
        enables public
    must_understand: bool | None
        enables must-understand
    no_transform: bool | None
        enables no-transform
    immutable: bool | None
        enables immutable
    stale_while_revalidate: int | None
        enables stale-while-revalidate, in seconds
    stale_if_error: int | None
        enables stale-if-error, in seconds

    For detailed information on Cache-Control header, please refer to:
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control
    """
    value = bytearray()

    if private and public:
        raise ValueError("A cache cannot be private and public at the same time.")

    def extend(part: bytes):
        if len(value):
            value.extend(b", ")
        value.extend(part)

    if max_age is not None:
        extend(f"max-age={max_age}".encode("ascii"))

    if shared_max_age is not None:
        extend(f"s-maxage={shared_max_age}".encode("ascii"))

    if no_cache:
        extend(b"no-cache")

    if must_revalidate:
        extend(b"must-revalidate")

    if proxy_revalidate:
        extend(b"proxy-revalidate")

    if no_store:
        extend(b"no-store")

    if private:
        extend(b"private")

    if public:
        extend(b"public")

    if must_understand:
        extend(b"must-understand")

    if no_transform:
        extend(b"no-transform")

    if immutable:
        extend(b"immutable")

    if stale_while_revalidate is not None:
        extend(f"stale-while-revalidate={stale_while_revalidate}".encode("ascii"))

    if stale_if_error is not None:
        extend(f"stale-if-error={stale_if_error}".encode("ascii"))

    return bytes(value)


class CacheControlHeaderValue:
    """
    Class used to represent a Cache-Control response header value.

    Parameters
    ----------
    max_age: int | None
        max age in seconds
    shared_max_age: int | None
        optional shared max age in seconds
    no_cache: bool | None
        enables no-cache
    no_store: bool | None
        enables no-store
    must_revalidate: bool | None
        enables must-revalidate
    proxy_revalidate: bool | None
        enables proxy-revalidate
    private: bool | None
        enables private
    public: bool | None
        enables public
    must_understand: bool | None
        enables must-understand
    no_transform: bool | None
        enables no-transform
    immutable: bool | None
        enables immutable
    stale_while_revalidate: int | None
        enables stale-while-revalidate, in seconds
    stale_if_error: int | None
        enables stale-if-error, in seconds

    For detailed information on Cache-Control header, please refer to:
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control
    """

    def __init__(
        self,
        *,
        max_age: Optional[int] = None,
        shared_max_age: Optional[int] = None,
        no_cache: Optional[bool] = None,
        no_store: Optional[bool] = None,
        must_revalidate: Optional[bool] = None,
        proxy_revalidate: Optional[bool] = None,
        private: Optional[bool] = None,
        public: Optional[bool] = None,
        must_understand: Optional[bool] = None,
        no_transform: Optional[bool] = None,
        immutable: Optional[bool] = None,
        stale_while_revalidate: Optional[int] = None,
        stale_if_error: Optional[int] = None,
    ) -> None:
        self.value: bytes = write_cache_control_response_header(
            max_age=max_age,
            shared_max_age=shared_max_age,
            no_cache=no_cache,
            no_store=no_store,
            must_revalidate=must_revalidate,
            proxy_revalidate=proxy_revalidate,
            private=private,
            public=public,
            must_understand=must_understand,
            no_transform=no_transform,
            immutable=immutable,
            stale_while_revalidate=stale_while_revalidate,
            stale_if_error=stale_if_error,
        )


def cache_control(
    max_age: Optional[int] = None,
    shared_max_age: Optional[int] = None,
    no_cache: Optional[bool] = None,
    no_store: Optional[bool] = None,
    must_revalidate: Optional[bool] = None,
    proxy_revalidate: Optional[bool] = None,
    private: Optional[bool] = None,
    public: Optional[bool] = None,
    must_understand: Optional[bool] = None,
    no_transform: Optional[bool] = None,
    immutable: Optional[bool] = None,
    stale_while_revalidate: Optional[int] = None,
    stale_if_error: Optional[int] = None,
):
    """
    Cache control decorator, applying a Cache-Control header to all successful
    responses.

    Parameters
    ----------
    max_age: int | None
        max age in seconds
    shared_max_age: int | None
        optional shared max age in seconds
    no_cache: bool | None
        enables no-cache
    no_store: bool | None
        enables no-store
    must_revalidate: bool | None
        enables must-revalidate
    proxy_revalidate: bool | None
        enables proxy-revalidate
    private: bool | None
        enables private
    public: bool | None
        enables public
    must_understand: bool | None
        enables must-understand
    no_transform: bool | None
        enables no-transform
    immutable: bool | None
        enables immutable
    stale_while_revalidate: int | None
        enables stale-while-revalidate, in seconds
    stale_if_error: int | None
        enables stale-if-error, in seconds

    For detailed information on Cache-Control header, please refer to:
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control
    """
    header_value: bytes = write_cache_control_response_header(
        max_age=max_age,
        shared_max_age=shared_max_age,
        no_cache=no_cache,
        no_store=no_store,
        must_revalidate=must_revalidate,
        proxy_revalidate=proxy_revalidate,
        private=private,
        public=public,
        must_understand=must_understand,
        no_transform=no_transform,
        immutable=immutable,
        stale_while_revalidate=stale_while_revalidate,
        stale_if_error=stale_if_error,
    )

    def decorator(next_handler):
        if inspect.iscoroutinefunction(next_handler):

            @wraps(next_handler)
            async def async_wrapped(*args, **kwargs):
                response = ensure_response(await next_handler(*args, **kwargs))
                response.add_header(b"cache-control", header_value)
                return response

            return async_wrapped
        else:

            @wraps(next_handler)
            def wrapped(*args, **kwargs):
                response = ensure_response(next_handler(*args, **kwargs))
                response.add_header(b"cache-control", header_value)
                return response

            copy_special_attributes(next_handler, wrapped)
            return wrapped

    return decorator


class CacheControlMiddleware:
    """
    The Cache-Control middleware lets configure a Cache-Control header globally, for
    all GET requests resulting in responses with status code 200 that don't have a
    cache-control header defined.

    Parameters
    ----------
    max_age: int | None
        max age in seconds
    shared_max_age: int | None
        optional shared max age in seconds
    no_cache: bool | None
        enables no-cache
    no_store: bool | None
        enables no-store
    must_revalidate: bool | None
        enables must-revalidate
    proxy_revalidate: bool | None
        enables proxy-revalidate
    private: bool | None
        enables private
    public: bool | None
        enables public
    must_understand: bool | None
        enables must-understand
    no_transform: bool | None
        enables no-transform
    immutable: bool | None
        enables immutable
    stale_while_revalidate: int | None
        enables stale-while-revalidate, in seconds
    stale_if_error: int | None
        enables stale-if-error, in seconds
    """

    def __init__(
        self,
        *,
        max_age: Optional[int] = None,
        shared_max_age: Optional[int] = None,
        no_cache: Optional[bool] = None,
        no_store: Optional[bool] = None,
        must_revalidate: Optional[bool] = None,
        proxy_revalidate: Optional[bool] = None,
        private: Optional[bool] = None,
        public: Optional[bool] = None,
        must_understand: Optional[bool] = None,
        no_transform: Optional[bool] = None,
        immutable: Optional[bool] = None,
        stale_while_revalidate: Optional[int] = None,
        stale_if_error: Optional[int] = None,
    ) -> None:
        self._header_value: bytes = write_cache_control_response_header(
            max_age=max_age,
            shared_max_age=shared_max_age,
            no_cache=no_cache,
            no_store=no_store,
            must_revalidate=must_revalidate,
            proxy_revalidate=proxy_revalidate,
            private=private,
            public=public,
            must_understand=must_understand,
            no_transform=no_transform,
            immutable=immutable,
            stale_while_revalidate=stale_while_revalidate,
            stale_if_error=stale_if_error,
        )

    def should_handle(self, request: Request, response: Response) -> bool:
        """
        Returns a value indicating whether the Cache-Control header should be set in
        the given response object, that was created for the given request.
        """
        return request.method == "GET" and response.status == 200

    async def __call__(self, request, handler):
        response = await handler(request)
        if (
            self.should_handle(request, response)
            and not response.headers[b"cache-control"]
        ):
            response.add_header(b"cache-control", self._header_value)

        return response



import logging
from contextlib import asynccontextmanager
from pathlib import Path
from typing import (
    Any,
    Awaitable,
    Callable,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

from guardpost.asynchronous.authentication import AuthenticationStrategy
from guardpost.asynchronous.authorization import AuthorizationStrategy
from guardpost.authorization import Policy, UnauthorizedError
from guardpost.common import AuthenticatedRequirement
from rodi import Container, Services

from blacksheep.baseapp import BaseApplication, handle_not_found
from blacksheep.common.files.asyncfs import FilesHandler
from blacksheep.contents import ASGIContent
from blacksheep.exceptions import HTTPException
from blacksheep.messages import Request, Response
from blacksheep.middlewares import get_middlewares_chain
from blacksheep.scribe import send_asgi_response
from blacksheep.server.asgi import get_request_url_from_scope
from blacksheep.server.authentication import (
    AuthenticateChallenge,
    get_authentication_middleware,
    handle_authentication_challenge,
)
from blacksheep.server.authorization import (
    AuthorizationWithoutAuthenticationError,
    get_authorization_middleware,
    handle_unauthorized,
)
from blacksheep.server.bindings import ControllerParameter
from blacksheep.server.controllers import router as controllers_router
from blacksheep.server.cors import CORSPolicy, CORSStrategy, get_cors_middleware
from blacksheep.server.env import EnvironmentSettings
from blacksheep.server.errors import ServerErrorDetailsHandler
from blacksheep.server.files import DefaultFileOptions, ServeFilesOptions
from blacksheep.server.files.dynamic import serve_files_dynamic
from blacksheep.server.normalization import normalize_handler, normalize_middleware
from blacksheep.server.responses import _ensure_bytes
from blacksheep.server.routing import (
    MountRegistry,
    RegisteredRoute,
    Router,
    RoutesRegistry,
)
from blacksheep.server.websocket import WebSocket
from blacksheep.sessions import Encryptor, SessionMiddleware, SessionSerializer, Signer
from blacksheep.utils import ensure_bytes, join_fragments

__all__ = ("Application",)


def get_default_headers_middleware(
    headers: Sequence[Tuple[str, str]],
) -> Callable[..., Awaitable[Response]]:
    raw_headers = tuple((name.encode(), value.encode()) for name, value in headers)

    async def default_headers_middleware(
        request: Request, handler: Callable[[Request], Awaitable[Response]]
    ) -> Response:
        response = await handler(request)
        for name, value in raw_headers:
            response.add_header(name, value)
        return response

    return default_headers_middleware


class ApplicationEvent:
    def __init__(self, context: Any) -> None:
        self._handlers: List[Callable[..., Any]] = []
        self.context = context

    def __iadd__(self, handler: Callable[..., Any]) -> "ApplicationEvent":
        self._handlers.append(handler)
        return self

    def __isub__(self, handler: Callable[..., Any]) -> "ApplicationEvent":
        self._handlers.remove(handler)
        return self

    def __len__(self) -> int:
        return len(self._handlers)

    def __call__(self, *args) -> Any:
        if args:
            self.__iadd__(args[0])
            return args[0]

        def decorator(fn):
            self.__iadd__(fn)
            return fn

        return decorator

    async def fire(self, *args: Any, **kwargs: Any) -> None:
        for handler in self._handlers:
            await handler(self.context, *args, **kwargs)


class ApplicationSyncEvent(ApplicationEvent):
    """
    ApplicationEvent whose subscribers must be synchronous functions.
    """

    def fire_sync(self, *args: Any, **keywargs: Any) -> None:
        for handler in self._handlers:
            handler(self.context, *args, **keywargs)

    async def fire(self, *args: Any, **keywargs: Any) -> None:
        raise TypeError(
            "The event handlers in this ApplicationEvent must be synchronous!"
        )


class ApplicationStartupError(RuntimeError):
    ...


class RequiresServiceContainerError(ApplicationStartupError):
    def __init__(self, details: str):
        super().__init__(
            f"The application requires services to be a Container "
            f"at this point of execution. Details: {details}"
        )
        self.details = details


class ApplicationAlreadyStartedCORSError(TypeError):
    def __init__(self) -> None:
        super().__init__(
            "The application is already running, configure CORS rules "
            "before starting the application"
        )


def _extend(obj, cls):
    """Applies a mixin to an instance of a class."""
    base_cls = obj.__class__
    base_cls_name = obj.__class__.__name__
    obj.__class__ = type(base_cls_name, (cls, base_cls), {})


class Application(BaseApplication):
    """
    Server application class.
    """

    def __init__(
        self,
        *,
        router: Optional[Router] = None,
        services: Optional[Container] = None,
        debug: bool = False,
        show_error_details: Optional[bool] = None,
        mount: Optional[MountRegistry] = None,
    ):
        env_settings = EnvironmentSettings()
        if router is None:
            router = Router()
        if services is None:
            services = Container()
        if show_error_details is None:
            show_error_details = env_settings.show_error_details
        if mount is None:
            mount = MountRegistry(env_settings.mount_auto_events)
        super().__init__(show_error_details, router)

        self.services: Container = services
        self._service_provider: Optional[Services] = None
        self.debug = debug
        self.middlewares: List[Callable[..., Awaitable[Response]]] = []
        self._default_headers: Optional[Tuple[Tuple[str, str], ...]] = None
        self._middlewares_configured = False
        self._cors_strategy: Optional[CORSStrategy] = None
        self._authentication_strategy: Optional[AuthenticationStrategy] = None
        self._authorization_strategy: Optional[AuthorizationStrategy] = None
        self.on_start = ApplicationEvent(self)
        self.after_start = ApplicationEvent(self)
        self.on_stop = ApplicationEvent(self)
        self.on_middlewares_configuration = ApplicationSyncEvent(self)
        self.started = False
        self.controllers_router: RoutesRegistry = controllers_router
        self.files_handler = FilesHandler()
        self.server_error_details_handler = ServerErrorDetailsHandler()
        self._session_middleware: Optional[SessionMiddleware] = None
        self.base_path: str = ""
        self._mount_registry = mount

    @property
    def service_provider(self) -> Services:
        """
        Returns the object that provides services of this application.
        """
        if self._service_provider is None:
            raise TypeError("The service provider is not build for this application.")
        return self._service_provider

    @property
    def default_headers(self) -> Optional[Tuple[Tuple[str, str], ...]]:
        return self._default_headers

    @default_headers.setter
    def default_headers(self, value: Optional[Tuple[Tuple[str, str], ...]]) -> None:
        self._default_headers = tuple(value) if value else None

    @property
    def cors(self) -> CORSStrategy:
        if not self._cors_strategy:
            raise TypeError(
                "CORS settings are not initialized for the application. "
                + "Use `app.use_cors()` method before using this property."
            )
        return self._cors_strategy

    @property
    def mount_registry(self) -> MountRegistry:
        return self._mount_registry

    def mount(self, path: str, app: Callable) -> None:
        """
        Mounts an ASGI application at the given path. When a web request has a URL path
        that starts with the mount path, it is handled by the mounted app (the mount
        path is stripped from the final URL path received by the child application).

        If the child application is a BlackSheep application, it requires handling of
        its lifecycle events. This can be automatic, if the environment variable

            APP_MOUNT_AUTO_EVENTS is set to "1" or "true" (case insensitive)

        or explicitly enabled, if the parent app's is configured this way:

            parent_app.mount_registry.auto_events = True
        """
        if app is self:
            raise TypeError("Cannot mount an application into itself")

        self._mount_registry.mount(path, app)

        if isinstance(app, Application):
            app.base_path = (
                join_fragments(self.base_path, path) if self.base_path else path
            )

            if self._mount_registry.auto_events:
                self._bind_child_app_events(app)

        if len(self._mount_registry.mounted_apps) == 1:
            # the first time a mount is configured, extend the application
            # to use mounts when handling web requests
            self.extend(MountMixin)

    def _bind_child_app_events(self, app: "Application") -> None:
        @self.on_start
        async def handle_child_app_start(_):
            await app.start()

        @self.after_start
        async def handle_child_app_after_start(_):
            await app.after_start.fire()

        @self.on_middlewares_configuration
        def handle_child_app_on_middlewares_configuration(_):
            app.on_middlewares_configuration.fire_sync()

        @self.on_stop
        async def handle_child_app_stop(_):
            await app.stop()

    def use_sessions(
        self,
        secret_key: str,
        *,
        session_cookie: str = "session",
        serializer: Optional[SessionSerializer] = None,
        signer: Optional[Signer] = None,
        encryptor: Optional[Encryptor] = None,
        session_max_age: Optional[int] = None,
    ) -> None:
        self._session_middleware = SessionMiddleware(
            secret_key=secret_key,
            session_cookie=session_cookie,
            serializer=serializer,
            signer=signer,
            encryptor=encryptor,
            session_max_age=session_max_age,
        )

    def use_cors(
        self,
        *,
        allow_methods: Union[None, str, Iterable[str]] = None,
        allow_headers: Union[None, str, Iterable[str]] = None,
        allow_origins: Union[None, str, Iterable[str]] = None,
        allow_credentials: bool = False,
        max_age: int = 5,
        expose_headers: Union[None, str, Iterable[str]] = None,
    ) -> CORSStrategy:
        """
        Enables CORS for the application, specifying the default rules to be applied
        for all request handlers.
        """
        if self.started:
            raise ApplicationAlreadyStartedCORSError()
        self._cors_strategy = CORSStrategy(
            CORSPolicy(
                allow_methods=allow_methods,
                allow_headers=allow_headers,
                allow_origins=allow_origins,
                allow_credentials=allow_credentials,
                max_age=max_age,
                expose_headers=expose_headers,
            ),
            self.router,
        )

        # Note: the following is a no-op request handler, necessary to activate handling
        # of OPTIONS preflight requests.
        # However, preflight requests are handled by the CORS middleware. This is to
        # stop the chain of middlewares and prevent extra logic from executing for
        # preflight requests (e.g. authentication logic)
        @self.router.options("*")
        async def options_handler(request: Request) -> Response:
            return Response(404)

        # User defined catch-all OPTIONS request handlers are not supported when the
        # built-in CORS handler is used.
        return self._cors_strategy

    def add_cors_policy(
        self,
        policy_name,
        *,
        allow_methods: Union[None, str, Iterable[str]] = None,
        allow_headers: Union[None, str, Iterable[str]] = None,
        allow_origins: Union[None, str, Iterable[str]] = None,
        allow_credentials: bool = False,
        max_age: int = 5,
        expose_headers: Union[None, str, Iterable[str]] = None,
    ) -> None:
        """
        Configures a set of CORS rules that can later be applied to specific request
        handlers, by name.

        The CORS policy can then be associated to specific request handlers,
        using the instance of `CORSStrategy` as a function decorator:

        @app.cors("example")
        @app.route("/")
        async def foo():
            ....
        """
        if self.started:
            raise ApplicationAlreadyStartedCORSError()

        if not self._cors_strategy:
            self.use_cors()

        assert self._cors_strategy is not None
        self._cors_strategy.add_policy(
            policy_name,
            CORSPolicy(
                allow_methods=allow_methods,
                allow_headers=allow_headers,
                allow_origins=allow_origins,
                allow_credentials=allow_credentials,
                max_age=max_age,
                expose_headers=expose_headers,
            ),
        )

    def use_authentication(
        self, strategy: Optional[AuthenticationStrategy] = None
    ) -> AuthenticationStrategy:
        if self.started:
            raise RuntimeError(
                "The application is already running, configure authentication "
                "before starting the application"
            )

        if self._authentication_strategy:
            return self._authentication_strategy

        if not strategy:
            strategy = AuthenticationStrategy()

        self._authentication_strategy = strategy
        return strategy

    def use_authorization(
        self, strategy: Optional[AuthorizationStrategy] = None
    ) -> AuthorizationStrategy:
        if self.started:
            raise RuntimeError(
                "The application is already running, configure authorization "
                "before starting the application"
            )

        if self._authorization_strategy:
            return self._authorization_strategy

        if not strategy:
            strategy = AuthorizationStrategy()

        if strategy.default_policy is None:
            # by default, a default policy is configured with no requirements,
            # meaning that request handlers allow anonymous users by default, unless
            # they are decorated with @auth()
            strategy.default_policy = Policy("default")
            strategy.add(Policy("authenticated").add(AuthenticatedRequirement()))

        self._authorization_strategy = strategy
        self.exceptions_handlers.update(
            {  # type: ignore
                AuthenticateChallenge: handle_authentication_challenge,
                UnauthorizedError: handle_unauthorized,
            }
        )
        return strategy

    def route(
        self, pattern: str, methods: Optional[Sequence[str]] = None
    ) -> Callable[..., Any]:
        if methods is None:
            methods = ["GET"]

        def decorator(f):
            for method in methods:
                self.router.add(method, pattern, f)
            return f

        return decorator

    def exception_handler(
        self, exception: Union[int, Type[Exception]]
    ) -> Callable[..., Any]:
        """
        Registers an exception handler function in the application exception handler.
        """

        def decorator(f):
            self.exceptions_handlers[exception] = f
            return f

        return decorator

    def lifespan(self, callback):
        """
        Registers an async generator, or async context manager, to be entered at
        application start, and exited at application shutdown. This is syntactic sugar
        alternative to handling application start and stop events directly. It can be
        useful to handle objects that need to be initialized and disposed, like HTTP
        clients that use connection pools, or files that needs to be open following the
        lifespan of the application.
        """
        if not hasattr(callback, "__aenter__"):
            callback = asynccontextmanager(callback)

        obj = None

        @self.on_start
        async def register_aenter(_):
            nonlocal obj
            try:
                obj = callback(self)
            except TypeError:
                obj = callback()
            await obj.__aenter__()

        @self.on_stop
        async def register_aexit(_):
            nonlocal obj
            if obj is not None:
                await obj.__aexit__(None, None, None)

        return callback

    def serve_files(
        self,
        source_folder: Union[str, Path],
        *,
        discovery: bool = False,
        cache_time: int = 10800,
        extensions: Optional[Set[str]] = None,
        root_path: str = "",
        index_document: Optional[str] = "index.html",
        fallback_document: Optional[str] = None,
        allow_anonymous: bool = True,
        default_file_options: Optional[DefaultFileOptions] = None,
    ):
        """
        Configures dynamic file serving from a given folder, relative to the server cwd.

        Parameters:
            source_folder (str): Path to the source folder containing static files.
            extensions: The set of files extensions to serve.
            discovery: Whether to enable file discovery, serving HTML pages for folders.
            cache_time: Controls the Cache-Control Max-Age in seconds for static files.
            root_path: Path prefix used for routing requests.
            For example, if set to "public", files are served at "/public/*".
            allow_anonymous: Whether to enable anonymous access to static files, true by
            default.
            index_document: The name of the index document to display, if present,
            in folders. Requests for folders that contain a file with matching produce
            a response with this document.
            fallback_document: Optional file name, for a document to serve when a
            response would be otherwise 404 Not Found; e.g. use this to serve SPA that
            use HTML5 History API for client side routing.
            default_file_options: Optional options to serve the default file
            (index.html)
        """
        if isinstance(source_folder, ServeFilesOptions):
            # deprecated class, will be removed in the next version
            from typing import cast

            deprecated_arg = cast(ServeFilesOptions, source_folder)
            deprecated_arg.validate()
            serve_files_dynamic(
                self.router,
                self.files_handler,
                str(deprecated_arg.source_folder),
                discovery=deprecated_arg.discovery,
                cache_time=deprecated_arg.cache_time,
                extensions=deprecated_arg.extensions,
                root_path=deprecated_arg.root_path,
                index_document=deprecated_arg.index_document,
                fallback_document=deprecated_arg.fallback_document,
                anonymous_access=deprecated_arg.allow_anonymous,
            )
            return
        serve_files_dynamic(
            self.router,
            self.files_handler,
            source_folder,
            discovery=discovery,
            cache_time=cache_time,
            extensions=extensions,
            root_path=root_path,
            index_document=index_document,
            fallback_document=fallback_document,
            anonymous_access=allow_anonymous,
            default_file_options=default_file_options,
        )

    def _apply_middlewares_in_routes(self):
        for route in self.router:
            route.handler = get_middlewares_chain(self.middlewares, route.handler)

    def _normalize_middlewares(self):
        self.middlewares = [
            normalize_middleware(middleware, self.service_provider)
            for middleware in self.middlewares
        ]

    def use_controllers(self):
        # NB: controller types are collected here, and not with
        # Controller.__subclasses__(),
        # to avoid funny bugs in case several Application objects are defined
        # with different controllers; this is the case for example of tests.

        # This sophisticated approach, using metaclassing, dynamic
        # attributes, and calling handlers dynamically
        # with activated instances of controllers; still supports custom
        # and generic decorators (*args, **kwargs);
        # as long as `functools.wraps` decorator is used in those decorators.
        self.register_controllers(self.prepare_controllers())

    def get_controller_handler_pattern(
        self, controller_type: Type, route: RegisteredRoute
    ) -> bytes:
        """
        Returns the full pattern to be used for a route handler,
        defined as controller method.
        """
        base_route = getattr(controller_type, "route", None)

        if base_route is not None:
            if callable(base_route):
                value = base_route()
            elif isinstance(base_route, (str, bytes)):
                value = base_route
            else:
                raise RuntimeError(
                    f"Invalid controller `route` attribute. "
                    f"Controller `{controller_type.__name__}` "
                    f"has an invalid route attribute: it should "
                    f"be callable, or str, or bytes."
                )

            if value:
                return ensure_bytes(join_fragments(value, route.pattern))
        return ensure_bytes(route.pattern)

    def prepare_controllers(self) -> List[Type]:
        controller_types = []
        for route in self.controllers_router:
            handler = route.handler
            controller_type = getattr(handler, "controller_type")
            controller_types.append(controller_type)
            handler.__annotations__["self"] = ControllerParameter[controller_type]
            self.router.add(
                route.method,
                self.get_controller_handler_pattern(controller_type, route),
                handler,
            )
        return controller_types

    def bind_controller_type(self, controller_type: Type):
        templates_environment = getattr(self, "templates_environment", None)

        if templates_environment:
            setattr(controller_type, "templates", templates_environment)

    def register_controllers(self, controller_types: List[Type]):
        """
        Registers controller types as transient services
        in the application service container.
        """
        if not controller_types:
            return

        if not isinstance(self.services, Container):
            raise RequiresServiceContainerError(
                "When using controllers, the application.services must be "
                "a service `Container` (`rodi.Container`; not a built service "
                "provider)."
            )

        for controller_class in controller_types:
            if controller_class in self.services:
                continue

            self.bind_controller_type(controller_class)

            if getattr(controller_class, "__init__") is object.__init__:
                self.services.add_transient_by_factory(
                    controller_class, controller_class
                )
            else:
                self.services.add_exact_transient(controller_class)

    def normalize_handlers(self):
        configured_handlers = set()

        self.router.sort_routes()

        for route in self.router:
            if route.handler in configured_handlers:
                continue

            route.handler = normalize_handler(route, self.service_provider)
            configured_handlers.add(route.handler)

        self._normalize_fallback_route()
        configured_handlers.clear()

    def _normalize_fallback_route(self):
        fallback = self.router.fallback

        if fallback is not None and self._has_default_not_found_handler():

            async def fallback_handler(app, request, exc) -> Response:
                return await fallback.handler(request)  # type: ignore

            self.exceptions_handlers[404] = fallback_handler  # type: ignore

    def _has_default_not_found_handler(self):
        return self.exceptions_handlers.get(404) is handle_not_found

    def configure_middlewares(self):
        if self._middlewares_configured:
            return
        self._middlewares_configured = True

        if self._authorization_strategy:
            if not self._authentication_strategy:
                raise AuthorizationWithoutAuthenticationError()
            self.middlewares.insert(
                0, get_authorization_middleware(self._authorization_strategy)
            )

        if self._authentication_strategy:
            self.middlewares.insert(
                0, get_authentication_middleware(self._authentication_strategy)
            )

        if self._session_middleware:
            self.middlewares.insert(0, self._session_middleware)

        if self._cors_strategy:
            self.middlewares.insert(0, get_cors_middleware(self, self._cors_strategy))

        if self._default_headers:
            self.middlewares.insert(
                0, get_default_headers_middleware(self._default_headers)
            )

        self.on_middlewares_configuration.fire_sync()

        self._normalize_middlewares()

        if self.middlewares:
            self._apply_middlewares_in_routes()

    def build_services(self):
        self._service_provider = self.services.build_provider()

    def extend(self, mixin) -> None:
        """
        Extends the class with additional features, applying the given mixin class.
        """
        _extend(self, mixin)

    async def start(self):
        if self.started:
            return

        self.started = True
        if self.on_start:
            await self.on_start.fire()

        self.use_controllers()
        self.build_services()
        self.normalize_handlers()
        self.configure_middlewares()

        if self.after_start:
            await self.after_start.fire()

    async def stop(self):
        await self.on_stop.fire()
        self.started = False

    async def _handle_lifespan(self, receive, send):
        message = await receive()
        assert message["type"] == "lifespan.startup"

        try:
            await self.start()
        except:  # NOQA
            logging.exception("Startup error")
            await send({"type": "lifespan.startup.failed"})
            return

        await send({"type": "lifespan.startup.complete"})

        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        await self.stop()
        await send({"type": "lifespan.shutdown.complete"})

    async def _handle_websocket(self, scope, receive, send):
        ws = WebSocket(scope, receive, send)
        route = self.router.get_ws_match(scope["path"])

        if route:
            ws.route_values = route.values
            try:
                return await route.handler(ws)
            except UnauthorizedError:
                await ws.close(401, "Unauthorized")
            except HTTPException as http_exception:
                await ws.close(http_exception.status, str(http_exception))
        await ws.close()

    async def _handle_http(self, scope, receive, send):
        assert scope["type"] == "http"

        request = Request.incoming(
            scope["method"],
            scope["raw_path"],
            scope["query_string"],
            list(scope["headers"]),
        )

        request.scope = scope
        request.content = ASGIContent(receive)

        response = await self.handle(request)
        await send_asgi_response(response, send)

        request.scope = None  # type: ignore
        request.content.dispose()

    async def __call__(self, scope, receive, send):
        if scope["type"] == "http":
            return await self._handle_http(scope, receive, send)

        if scope["type"] == "websocket":
            return await self._handle_websocket(scope, receive, send)

        if scope["type"] == "lifespan":
            return await self._handle_lifespan(receive, send)

        raise TypeError(f"Unsupported scope type: {scope['type']}")


class MountMixin:
    _mount: MountRegistry
    base_path: str

    def handle_mount_path(self, scope, route_match):
        assert route_match.values is not None
        tail = route_match.values.get("tail")
        assert tail is not None
        tail = "/" + tail

        scope["path"] = tail
        scope["raw_path"] = tail.encode("utf8")

    async def _handle_redirect_to_mount_root(self, scope, send):
        """
        A request to the path "https://.../{mount_path}" must result in a
        307 Temporary Redirect to the root of the mount: "https://.../{mount_path}/"
        including a trailing slash.
        """
        response = Response(
            307,
            [
                (
                    b"Location",
                    _ensure_bytes(
                        get_request_url_from_scope(
                            scope, trailing_slash=True, base_path=self.base_path
                        )
                    ),
                )
            ],
        )
        await send_asgi_response(response, send)

    async def __call__(self, scope, receive, send):
        if scope["type"] == "lifespan":
            return await super()._handle_lifespan(receive, send)  # type: ignore

        for route in self.mount_registry.mounted_apps:  # type: ignore
            route_match = route.match(scope["raw_path"])
            if route_match:
                raw_path = scope["raw_path"]
                if raw_path == route.pattern.rstrip(b"/*") and scope["type"] == "http":
                    return await self._handle_redirect_to_mount_root(scope, send)
                self.handle_mount_path(scope, route_match)
                return await route.handler(scope, receive, send)

        return await super().__call__(scope, receive, send)  # type: ignore



import asyncio
import ssl
from asyncio import AbstractEventLoop, TimeoutError
from typing import (
    Any,
    AnyStr,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from urllib.parse import urlencode

from blacksheep import URL, Content, InvalidURL, Request, Response
from blacksheep.middlewares import get_middlewares_chain
from blacksheep.utils.aio import get_running_loop

from .connection import ConnectionClosedError
from .cookies import CookieJar, cookies_middleware
from .exceptions import (
    CircularRedirectError,
    ConnectionTimeout,
    MaximumRedirectsExceededError,
    MissingLocationForRedirect,
    RequestTimeout,
    UnsupportedRedirect,
)
from .pool import ClientConnection, ConnectionPools

URLType = Union[str, bytes, URL]
KeyValuePair = Tuple[AnyStr, AnyStr]
Headers = Union[Dict[AnyStr, AnyStr], Iterable[KeyValuePair]]
Params = Union[Dict[AnyStr, AnyStr], Iterable[KeyValuePair]]


def _ensure_header_bytes(value: AnyStr) -> bytes:
    return value if isinstance(value, bytes) else value.encode("ascii")


def normalize_headers(
    headers: Optional[Headers],
) -> Optional[List[Tuple[bytes, bytes]]]:
    if headers is None:
        return None
    if isinstance(headers, dict):
        return [
            (_ensure_header_bytes(key), _ensure_header_bytes(value))
            for key, value in headers.items()
        ]
    return [
        (_ensure_header_bytes(key), _ensure_header_bytes(value))
        for key, value in headers
    ]


class RedirectsCache:
    """Used to store permanent redirects urls for later reuse"""

    __slots__ = ("_cache",)

    def __init__(self):
        self._cache: Dict[bytes, URL] = {}

    def __setitem__(self, key: bytes, value: URL):
        self._cache[key] = value

    def __getitem__(self, item: Any) -> Optional[URL]:
        try:
            return self._cache[item]
        except KeyError:
            return None

    def __contains__(self, item: Any) -> bool:
        return item in self._cache


class ClientRequestContext:
    __slots__ = ("path", "cookies")

    def __init__(self, request, cookies: Optional[CookieJar] = None):
        self.path = [request.url.value.lower()]
        self.cookies = cookies


_default_pools_by_loop_id = {}


def _get_default_pools_for_loop(loop):
    """
    This function is meant to help users of the ClientSession:
    to prevent creating many pools, decreasing performance only by
    instantiating several ClientSession without a specific HTTP Connections
    pools.
    """
    loop_id = id(loop)
    try:
        return _default_pools_by_loop_id[loop_id]
    except KeyError:
        pools = ConnectionPools(loop)
        _default_pools_by_loop_id[loop_id] = pools
        return pools


class ClientSession:
    def __init__(
        self,
        loop: Optional[AbstractEventLoop] = None,
        base_url: Union[None, bytes, str, URL] = None,
        ssl: Union[None, bool, ssl.SSLContext] = None,
        pools: Optional[ConnectionPools] = None,
        default_headers: Optional[Headers] = None,
        follow_redirects: bool = True,
        connection_timeout: float = 10.0,
        request_timeout: float = 60.0,
        maximum_redirects: int = 20,
        redirects_cache_type: Union[Type[RedirectsCache], Any] = None,
        cookie_jar: Union[None, bool, CookieJar] = None,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        if loop is None:
            loop = get_running_loop()

        if not pools:
            # TODO: maybe it is safer to emit a warning here,
            # and create an instance of ClientConnectionPools
            pools = _get_default_pools_for_loop(loop)
            self.owns_pools = False
        else:
            self.owns_pools = True

        if redirects_cache_type is None and follow_redirects:
            redirects_cache_type = RedirectsCache

        if middlewares is None:
            middlewares = []

        if cookie_jar is None:
            cookie_jar = CookieJar()

        if cookie_jar is False:
            cookie_jar = None
        else:
            middlewares.insert(0, cookies_middleware)

        self.loop = loop
        self._base_url: Optional[URL]
        self.base_url = base_url
        self.ssl = ssl
        self._default_headers: Optional[List[Tuple[bytes, bytes]]]
        self.default_headers = default_headers
        self.pools = pools
        self.connection_timeout = connection_timeout
        self.request_timeout = request_timeout
        self.follow_redirects = follow_redirects
        self.cookie_jar: Optional[CookieJar] = cast(Optional[CookieJar], cookie_jar)
        self._permanent_redirects_urls = (
            redirects_cache_type() if follow_redirects else None
        )
        self.non_standard_handling_of_301_302_redirect_method = True
        self.maximum_redirects = maximum_redirects
        self._handler = None
        self._middlewares: List[Callable[..., Any]]
        self.middlewares = middlewares
        self.delay_before_retry = 0.5

    @property
    def default_headers(self) -> Optional[List[Tuple[bytes, bytes]]]:
        return self._default_headers

    @default_headers.setter
    def default_headers(self, value: Optional[Headers]):
        self._default_headers = normalize_headers(value)

    @property
    def middlewares(self) -> List[Callable[..., Any]]:
        return self._middlewares

    @middlewares.setter
    def middlewares(self, value) -> None:
        self._middlewares = list(value)
        self._build_middlewares_chain()

    @property
    def base_url(self) -> Optional[URL]:
        return self._base_url

    @base_url.setter
    def base_url(self, value):
        url = None
        if value and not isinstance(value, URL):
            if isinstance(value, str):
                url = URL(value.encode())
            else:
                url = URL(value)
        self._base_url = url

    def add_middlewares(self, middlewares: List[Callable]):
        self._middlewares += middlewares
        self._build_middlewares_chain()

    def _build_middlewares_chain(self):
        if not self.middlewares:
            return

        async def root_handler(request):
            return await self._send_core(request)

        self._handler = get_middlewares_chain(self._middlewares, root_handler)

    def use_standard_redirect(self):
        """Uses specification compliant handling of 301 and 302 redirects"""
        self.non_standard_handling_of_301_302_redirect_method = False

    def get_url(self, url, params: Optional[Params] = None) -> bytes:
        value = self.get_url_value(url)
        if not params:
            return value
        query = urlencode(params).encode("ascii")
        return value + (b"&" if b"?" in value else b"?") + query

    def get_url_value(self, url: Union[AnyStr, URL]) -> bytes:
        if isinstance(url, str):
            url = url.encode()

        if not isinstance(url, URL):
            if url == b"":
                url = b"/"
            url = URL(url)

        if url.is_absolute:
            return url.value

        if self.base_url:
            return self.base_url.join(url).value
        return url.value

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        if self.owns_pools:
            self.pools.dispose()

    @staticmethod
    def extract_redirect_location(response: Response) -> URL:
        # if the server returned more than one value, use
        # the first header in order
        location = response.get_first_header(b"Location")
        if not location:
            raise MissingLocationForRedirect(response)

        # if the location cannot be parsed as URL, let exception happen:
        # this might be a redirect to a URN!
        # simply don't follows the redirect, and returns the response to
        # the caller
        try:
            return URL(location)
        except InvalidURL:
            raise UnsupportedRedirect(location)

    @staticmethod
    def get_redirect_url(request: Request, location: URL) -> URL:
        if location.is_absolute:
            return location
        # relative redirect URI
        # https://tools.ietf.org/html/rfc7231#section-7.1.2
        return request.url.base_url().join(location)

    def validate_redirect(
        self, redirect_url: URL, response: Response, context: ClientRequestContext
    ):
        redirect_url_lower = redirect_url.value.lower()
        if redirect_url_lower in context.path:
            context.path.append(redirect_url_lower)
            raise CircularRedirectError(context.path, response)

        context.path.append(redirect_url_lower)

        if len(context.path) > self.maximum_redirects:
            raise MaximumRedirectsExceededError(
                context.path, response, self.maximum_redirects
            )

    def update_request_for_redirect(self, request: Request, response: Response) -> None:
        context: ClientRequestContext = getattr(
            request, "context", ClientRequestContext(request, self.cookie_jar)
        )
        status = response.status

        if status == 301 or status == 302:
            if (
                self.non_standard_handling_of_301_302_redirect_method
                and request.method != "GET"
            ):
                # Change original request method to GET (Browser-like)
                request.method = "GET"

        if status == 303:
            # 303 See Other
            # Change original request method to GET
            request.method = "GET"

        location = self.extract_redirect_location(response)
        redirect_url = self.get_redirect_url(request, location)

        self.validate_redirect(redirect_url, response, context)

        if status == 301 or status == 308:
            self._permanent_redirects_urls[request.url.value] = redirect_url

        request.url = redirect_url

    def merge_default_headers(self, request: Request) -> None:
        if not self.default_headers:
            return

        for header in self.default_headers:
            if header[0] not in request.headers:
                request.headers.add(header[0], header[1])

    def check_permanent_redirects(self, request: Request) -> None:
        if (
            self._permanent_redirects_urls is not None
            and request.url.value in self._permanent_redirects_urls
        ):
            redirect_url = self._permanent_redirects_urls[request.url.value]
            if redirect_url is not None:
                request.url = redirect_url

    async def get_connection(self, url: URL) -> ClientConnection:
        pool = self.pools.get_pool(url.schema, url.host, url.port, self.ssl)

        try:
            return await asyncio.wait_for(
                pool.get_connection(), self.connection_timeout
            )
        except TimeoutError:
            raise ConnectionTimeout(url.base_url(), self.connection_timeout)

    def get_new_context(self, request: Request) -> ClientRequestContext:
        return ClientRequestContext(request, self.cookie_jar)

    def _validate_request_url(self, request: Request):
        if not request.url.is_absolute:
            if self.base_url:
                request.url = URL(self.get_url_value(request.url))
            else:
                raise ValueError(
                    "request.url must be a complete, absolute URL. "
                    "Either provide a base_url "
                    "for the client, or specify a full URL for the request."
                )

    async def send(self, request: Request) -> Response:
        self._validate_request_url(request)

        if not hasattr(request, "context"):
            request.context = self.get_new_context(request)  # type: ignore
            self.merge_default_headers(request)

        if self._handler:
            # using middlewares
            response = await self._handler(request)
        else:
            # without middlewares
            response = await self._send_core(request)

        if self.follow_redirects and response.is_redirect():
            try:
                self.update_request_for_redirect(request, response)
            except UnsupportedRedirect:
                # redirect not to HTTP / HTTPS: for example,
                # it can be a redirect to a URN
                return response
            return await self.send(request)

        return response

    async def _send_core(self, request: Request) -> Response:
        self.check_permanent_redirects(request)

        return await self._send_using_connection(request)

    async def _send_using_connection(self, request, attempt: int = 1) -> Response:
        connection = await self.get_connection(request.url)

        try:
            return await asyncio.wait_for(
                connection.send(request), self.request_timeout
            )
        except ConnectionClosedError as connection_closed_error:
            if connection_closed_error.can_retry and attempt < 4:
                await asyncio.sleep(self.delay_before_retry)
                return await self._send_using_connection(request, attempt + 1)
            raise
        except TimeoutError:
            raise RequestTimeout(request.url, self.request_timeout)

    async def get(
        self,
        url: URLType,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        return await self.send(
            Request("GET", self.get_url(url, params), normalize_headers(headers))
        )

    async def post(
        self,
        url: URLType,
        content: Optional[Content] = None,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        request = Request("POST", self.get_url(url, params), normalize_headers(headers))
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def put(
        self,
        url: URLType,
        content: Optional[Content] = None,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        request = Request("PUT", self.get_url(url, params), normalize_headers(headers))
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def delete(
        self,
        url: URLType,
        content: Optional[Content] = None,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        request = Request(
            "DELETE", self.get_url(url, params), normalize_headers(headers)
        )
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def trace(
        self,
        url: URLType,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        return await self.send(
            Request("TRACE", self.get_url(url, params), normalize_headers(headers))
        )

    async def head(
        self,
        url: URLType,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        return await self.send(
            Request("HEAD", self.get_url(url, params), normalize_headers(headers))
        )

    async def patch(
        self,
        url: URLType,
        content: Optional[Content] = None,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        request = Request(
            "PATCH", self.get_url(url, params), normalize_headers(headers)
        )
        return await self.send(
            request.with_content(content) if content is not None else request
        )

    async def options(
        self,
        url: URLType,
        content: Optional[Content] = None,
        headers: Optional[Headers] = None,
        params: Optional[Params] = None,
    ) -> Response:
        request = Request(
            "OPTIONS", self.get_url(url, params), normalize_headers(headers)
        )
        return await self.send(
            request.with_content(content) if content is not None else request
        )



import re

import pytest
from jinja2 import Environment, PackageLoader

from blacksheep.contents import write_www_form_urlencoded
from blacksheep.messages import Response
from blacksheep.server.controllers import Controller
from blacksheep.server.csrf import (
    AntiForgeryBaseExtension,
    ignore_anti_forgery,
    use_anti_forgery,
)
from blacksheep.server.responses import no_content
from blacksheep.server.routing import RoutesRegistry
from blacksheep.server.templating import use_templates, view, view_async
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


def read_control_value_from_input(text: str) -> str:
    match = re.search(r'name="__RequestVerificationToken" value="([^\"]+)"', text)

    assert match is not None
    return match.group(1)


def get_app(enable_async=False):
    app = FakeApplication()
    render = use_templates(
        app,
        loader=PackageLoader("tests.testapp", "templates"),
        enable_async=enable_async,
    )

    use_anti_forgery(app)

    return app, render


@pytest.fixture()
def home_model():
    return {
        "title": "Example",
        "heading": "Hello World!",
        "paragraph": "Lorem ipsum dolor sit amet",
    }


async def _assert_generation_scenario(response: Response):
    text = await response.text()

    assert '<input type="hidden" name="__RequestVerificationToken" value=' in text

    assert response.status == 200

    af_cookie = response.cookies["aftoken"]
    assert af_cookie is not None

    value = af_cookie.value
    assert value is not None

    # the value in the cookie and the value in the input element must be different
    control_value = read_control_value_from_input(text)
    assert value != control_value


@pytest.mark.asyncio
async def test_anti_forgery_token_generation(home_model):
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None
    await _assert_generation_scenario(app.response)


@pytest.mark.asyncio
async def test_anti_forgery_token_click_jacking_protection(home_model):
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    csp = app.response.headers[b"Content-Security-Policy"]
    assert csp == (b"frame-ancestors: 'self';",)

    x_frame_options = app.response.headers[b"X-Frame-Options"]
    assert x_frame_options == (b"SAMEORIGIN",)


@pytest.mark.asyncio
async def test_anti_forgery_token_generation_multiple(home_model):
    """
    Verifies that using {% af_input %} doesn't generate multiple values for the same
    web request, and that {% af_token %} returns the token without an HTML fragment.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_2", home_model, request=request)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    response = app.response
    assert response is not None

    text = await response.text()

    values = list(
        re.findall(r'name="__RequestVerificationToken" value="([^\"]+)"', text)
    )

    assert len(values) == 2

    # all values must be equal
    assert all(value == values[0] for value in values)

    match = re.search(r'{"token": "([^\"]+)"}', text)

    assert match is not None
    value = match.group(1)
    assert value == values[0]

    # the response cookie is only one
    af_cookie = response.cookies["aftoken"]
    assert af_cookie is not None


@pytest.mark.asyncio
async def test_anti_forgery_missing_request_context(home_model):
    app, render = get_app(False)

    @app.router.get("/")
    async def home():
        return render("form_1", home_model)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None
    assert app.response.status == 500

    text = await app.response.text()
    assert text is not None
    assert (
        "blacksheep.server.csrf.MissingRequestContextError: The request context is "
        "missing from the render call. Pass the request object to the context of the "
        "template."
    ) in text


async def _valid_scenario(app: FakeApplication):
    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    text = await app.response.text()

    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None

    # a valid request includes both the control value and the cookie:
    # extract the anti-forgery token, then do POST
    control_value = read_control_value_from_input(text)
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
            cookies={"aftoken": af_cookie.value},
        ),
        MockReceive(
            [
                write_www_form_urlencoded(
                    {
                        "username": "Charlie Brown",
                        "__RequestVerificationToken": control_value,
                    }
                )
            ]
        ),
        MockSend(),
    )

    response = app.response
    assert response.status == 204


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_input_1(home_model):
    """
    Tests a valid scenario.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await _valid_scenario(app)


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_input_1_view_fn(home_model):
    """
    Tests a valid scenario using responses.view function.
    """
    app, _ = get_app(False)

    @app.router.get("/")
    async def home(env: Environment, request):
        return view(env, "form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await _valid_scenario(app)


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_input_1_async_view_fn(home_model):
    """
    Tests a valid scenario using responses.view function.
    """
    app, _ = get_app(True)

    @app.router.get("/")
    async def home(env: Environment, request):
        return await view_async(env, "form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await _valid_scenario(app)


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_input_1b(home_model):
    """
    Tests handling of an anomalous situation when the same form contains two elements
    for the verification.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    text = await app.response.text()

    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None

    # a valid request includes both the control value and the cookie:
    # extract the anti-forgery token, then do POST
    control_value = read_control_value_from_input(text).encode()
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
            cookies={"aftoken": af_cookie.value},
        ),
        MockReceive(
            [
                b"username=Charlie+Brown"
                + b"&__RequestVerificationToken="
                + control_value
                + b"&__RequestVerificationToken="
                + control_value
            ]
        ),
        MockSend(),
    )

    response = app.response
    assert response.status == 204


@pytest.mark.asyncio
async def test_anti_forgery_cookie_must_be_http_only(home_model):
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None
    assert af_cookie.http_only is True


@pytest.mark.asyncio
async def test_tokens_reuse_across_requests(home_model):
    """
    When the same client has multiple pages open on the same website, they should use
    by default the same cookie (otherwise opening a new tab breaks the previous tabs
    since the control values in input elements would not be updated!)
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    first_cookie = app.response.cookies["aftoken"]
    assert first_cookie is not None

    # now get the same page, but with a request cookie that already contains an
    # anti-forgery token that was generated previously (this could happen in the same
    # page or in another page of the same website)
    await app(
        get_example_scope("GET", "/", cookies={"aftoken": first_cookie.value}),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None

    # the second response doesn't have a set-cookie header because it's not necessary:
    # the client already has a cookie for anti-forgery validation
    # but we can make a web request with a new control value generated for the second
    # response
    second_cookie = app.response.cookies.get("aftoken")
    assert second_cookie is None

    text = await app.response.text()
    control_value = read_control_value_from_input(text)
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
            cookies={"aftoken": first_cookie.value},
        ),
        MockReceive(
            [
                write_www_form_urlencoded(
                    {
                        "username": "Charlie Brown",
                        "__RequestVerificationToken": control_value,
                    }
                )
            ]
        ),
        MockSend(),
    )

    response = app.response
    assert response.status == 204


@pytest.mark.asyncio
async def test_tokens_reuse_across_requests_invalid_token(home_model):
    """
    Tests an invalid value sent for aftoken cookie.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    # now get the same page, but with a request cookie that already contains an
    # anti-forgery token that was generated previously (this could happen in the same
    # page or in another page of the same website)
    await app(
        get_example_scope("GET", "/", cookies={"aftoken": "invalid value"}),
        MockReceive(),
        MockSend(),
    )

    assert app.response is not None

    # the second response doesn't have a set-cookie header because it's not necessary:
    # the client already has a cookie for anti-forgery validation
    # but we can make a web request with a new control value generated for the second
    # response
    second_cookie = app.response.cookies.get("aftoken")
    assert second_cookie is not None


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_input_2(home_model):
    """
    Tests invalid request, missing the value in the cookie.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    text = await app.response.text()

    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None

    # a valid request includes both the control value and the cookie:
    # extract the anti-forgery token, then do POST
    control_value = read_control_value_from_input(text)

    # a request without cookie will fail:
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
        ),
        MockReceive(
            [
                write_www_form_urlencoded(
                    {
                        "username": "Charlie Brown",
                        "__RequestVerificationToken": control_value,
                    }
                )
            ]
        ),
        MockSend(),
    )

    response = app.response
    assert response.status == 401
    reason = response.headers[b"Reason"]
    assert reason == (b"Missing anti-forgery token cookie",)


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_input_3(home_model):
    """
    Tests invalid control value, that cannot be deserialized (BadSignature).
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None

    # a valid request includes both the control value and the cookie:
    # extract the anti-forgery token, then do POST

    # a request with invalid control value will fail:
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
            cookies={"aftoken": af_cookie.value},
        ),
        MockReceive(
            [
                write_www_form_urlencoded(
                    {
                        "username": "Charlie Brown",
                        "__RequestVerificationToken": "This is wrong",
                    }
                )
            ]
        ),
        MockSend(),
    )

    response = app.response
    assert response.status == 401
    reason = response.headers[b"Reason"]
    assert reason == (b"Invalid anti-forgery token",)


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_input_4(home_model):
    """
    Tests invalid control value, which was generated for another request.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    text = await app.response.text()
    control_value = read_control_value_from_input(text)
    # a valid request includes both the control value and the cookie:
    # extract the anti-forgery token, then do POST

    # make another web request, which results in obtaining a new unrelated AF token,
    # then use the cookie AF token with the unrelated control value
    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None

    # a request with invalid control value will fail:
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
            cookies={"aftoken": af_cookie.value},
        ),
        MockReceive(
            [
                write_www_form_urlencoded(
                    {
                        "username": "Charlie Brown",
                        "__RequestVerificationToken": control_value,
                    }
                )
            ]
        ),
        MockSend(),
    )

    response = app.response
    assert response.status == 401
    reason = response.headers[b"Reason"]
    assert reason == (b"Invalid anti-forgery token",)


@pytest.mark.asyncio
async def test_missing_control_value(home_model):
    """
    Tests missing control value, which was generated for another request.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None

    # a request with invalid control value will fail:
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
            cookies={"aftoken": af_cookie.value},
        ),
        MockReceive([write_www_form_urlencoded({"username": "Charlie Brown"})]),
        MockSend(),
    )

    response = app.response
    assert response.status == 401
    reason = response.headers[b"Reason"]
    assert reason == (b"Missing anti-forgery token control value",)


@pytest.mark.asyncio
async def test_anti_forgery_token_validation_using_header(home_model):
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    text = await app.response.text()

    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None

    # a valid request includes both the control value and the cookie:
    # extract the anti-forgery token, then do POST
    control_value = read_control_value_from_input(text)
    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "RequestVerificationToken": control_value,
            },
            cookies={"aftoken": af_cookie.value},
        ),
        MockReceive([write_www_form_urlencoded({"username": "Charlie Brown"})]),
        MockSend(),
    )

    response = app.response
    assert response.status == 204


@pytest.mark.asyncio
async def test_controller_view_generation(home_model):
    app, _ = get_app(False)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get("/")
        def form_1(self, request):
            return self.view("form_1", model=home_model, request=request)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None
    await _assert_generation_scenario(app.response)


@pytest.mark.asyncio
async def test_controller_async_view_generation(home_model):
    app, _ = get_app(True)
    app.controllers_router = RoutesRegistry()
    get = app.controllers_router.get

    class Lorem(Controller):
        @get("/")
        async def form_1(self, request):
            return await self.view_async("form_1", model=home_model, request=request)

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None
    await _assert_generation_scenario(app.response)


@pytest.mark.asyncio
async def test_anti_forgery_base_extension_raises_without_handler(home_model):
    app, _ = get_app(False)

    env = app.templates_environment  # type: ignore

    with pytest.raises(TypeError):
        AntiForgeryBaseExtension(env)

    class Foo(AntiForgeryBaseExtension):
        af_handler = object()  # type: ignore

    instance = Foo(env)
    instance.af_handler = None

    with pytest.raises(TypeError):
        instance.handler


@pytest.mark.asyncio
async def test_anti_forgery_ignore_decorator(home_model):
    """
    Tests a valid scenario, using the ignore_anti_forgery decorator.
    """
    app, render = get_app(False)

    @app.router.get("/")
    async def home(request):
        return render("form_1", home_model, request=request)

    @ignore_anti_forgery()
    @app.router.post("/user")
    async def create_username():
        return no_content()

    await app.start()

    await app(get_example_scope("GET", "/"), MockReceive(), MockSend())
    assert app.response is not None

    assert app.response is not None
    af_cookie = app.response.cookies["aftoken"]
    assert af_cookie is not None, "An Anti-Forgery token is still issued for the home."

    await app(
        get_example_scope(
            "POST",
            "/user",
            extra_headers={"Content-Type": "application/x-www-form-urlencoded"},
        ),
        MockReceive([write_www_form_urlencoded({"username": "Charlie Brown"})]),
        MockSend(),
    )

    response = app.response
    assert response.status == 204



import os
from email.utils import formatdate

from .pathsutils import get_mime_type_from_name


class FileInfo:
    __slots__ = ("etag", "size", "mime", "modified_time")

    def __init__(self, size: int, etag: str, mime: str, modified_time: str):
        self.size = size
        self.etag = etag
        self.mime = mime
        self.modified_time = modified_time

    def __repr__(self):
        return (
            f"<FileInfo mime={self.mime} "
            f"etag={self.etag} "
            f"modified_time={self.modified_time}>"
        )

    def to_dict(self):
        return {key: getattr(self, key, None) for key in FileInfo.__slots__}

    @classmethod
    def from_path(cls, resource_path: str):
        stat = os.stat(resource_path)
        return cls(
            stat.st_size,
            str(stat.st_mtime),
            get_mime_type_from_name(resource_path),
            formatdate(stat.st_mtime, usegmt=True),
        )



# cython: language_level=3, embedsignature=True
# Copyright (C) 2018-present Roberto Prevato
#
# This module is part of BlackSheep and is released under
# the MIT License https://opensource.org/licenses/MIT


cdef class Header:
    cdef readonly bytes name
    cdef readonly bytes value


cdef class Headers:
    cdef readonly list values

    cpdef tuple keys(self)

    cpdef Headers clone(self)

    cpdef tuple get(self, bytes name)

    cpdef list get_tuples(self, bytes name)

    cpdef void add(self, bytes name, bytes value)

    cpdef void set(self, bytes name, bytes value)

    cpdef bytes get_single(self, bytes name)

    cpdef bytes get_first(self, bytes name)

    cpdef void remove(self, bytes key)

    cpdef void merge(self, list values)

    cpdef bint contains(self, bytes key)



import uuid
from typing import Any, AsyncIterable, Callable, Dict, List, Optional, Tuple, Union

class Content:
    def __init__(self, content_type: bytes, data: bytes):
        self.type = content_type
        self.body = data
        self.length = len(data)
    async def read(self) -> bytes:
        return self.body
    def dispose(self) -> None: ...

class StreamedContent(Content):
    def __init__(
        self,
        content_type: bytes,
        data_provider: Callable[[], AsyncIterable[bytes]],
    ) -> None:
        self.type = content_type
        self.body = None
        self.length = -1
        self.generator = data_provider
    async def get_parts(self) -> AsyncIterable[bytes]: ...

class ASGIContent(Content):
    def __init__(self, receive: Callable[[], bytes]):
        self.type = None
        self.body = None
        self.length = -1
        self.receive = receive
    def dispose(self): ...
    async def stream(self) -> AsyncIterable[bytes]: ...
    async def read(self) -> bytes: ...

class TextContent(Content):
    def __init__(self, text: str):
        super().__init__(b"text/plain; charset=utf-8", text.encode("utf8"))

class HTMLContent(Content):
    def __init__(self, html: str):
        super().__init__(b"text/html; charset=utf-8", html.encode("utf8"))

def default_json_dumps(value: Any) -> str: ...

class JSONContent(Content):
    def __init__(self, data: object, dumps: Callable[[Any], str] = default_json_dumps):
        """
        Creates an instance of JSONContent class, automatically serializing the given
        input in JSON format, encoded using UTF-8.
        """
        super().__init__(b"application/json", dumps(data).encode("utf8"))

class FormContent(Content):
    def __init__(self, data: Union[Dict[str, str], List[Tuple[str, str]]]):
        """
        Creates an instance of FormContent class, with application/x-www-form-urlencoded
        type, and bytes data serialized from the given dictionary.

        :param data: data to be serialized.
        """
        super().__init__(b"application/x-www-form-urlencoded", b"")

class FormPart:
    def __init__(
        self,
        name: bytes,
        data: bytes,
        content_type: Optional[bytes] = None,
        file_name: Optional[bytes] = None,
        charset: Optional[bytes] = None,
    ):
        self.name = name
        self.data = data
        self.file_name = file_name
        self.content_type = content_type
        self.charset = charset
    def __repr__(self):
        return f"<FormPart {self.name} - at {id(self)}>"

class MultiPartFormData(Content):
    def __init__(self, parts: List[FormPart]):
        self.parts = parts
        self.boundary = b"------" + str(uuid.uuid4()).replace("-", "").encode()
        super().__init__(b"multipart/form-data; boundary=" + self.boundary, b"")

def parse_www_form(content: str) -> Dict[str, Union[str, List[str]]]:
    """Parses application/x-www-form-urlencoded content"""

def write_www_form_urlencoded(
    data: Union[Dict[str, str], List[Tuple[str, str]]]
) -> bytes: ...

HtmlContent = HTMLContent  # for backward compatibility
JsonContent = JSONContent  # for backward compatibility
